/* Copyright (c) 2002-2006, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

/*
 * vapi_const.h 
 * 
 * This file contains "tunable parameters" for VAPI. A lot of work 
 * needs to be done to find out what are the right values, and in 
 * fact whether this is the right set of knobs. Many are settable 
 * from the command line and
 * passed to the individual processes as environment variables.
 */

#ifndef _IBVERBS_CONST_H
#define _IBVERBS_CONST_H

/*#define CO_PRINT */
#ifdef CO_PRINT
#define co_print(arg...)                    \
do {                                        \
   printf("[%d]", viadev.me);               \
   printf(arg);                             \
} while (0)
#else
#define co_print(arg...)
#endif

#define VIADEV_DEFAULT_DEVICE                 "mthca0"
#define VIADEV_INVALID_DEVICE                 "nohca"

#define QPLEN_XDR                             (8)

#define VIADEV_MAX_RDMA_SIZE		          (4194304)  
#define VIADEV_DEFAULT_QP_OUS_RD_ATOM         (4)

#define VIADEV_DEFAULT_PSN                    (0)
#define VIADEV_DEFAULT_PKEY_IX                (0) 
#define VIADEV_DEFAULT_P_KEY                  (0x0)
#define VIADEV_DEFAULT_MIN_RNR_TIMER          (12)
#define VIADEV_DEFAULT_SERVICE_LEVEL          (0)
#define VIADEV_DEFAULT_TIME_OUT               (10)

#define VIADEV_DEFAULT_STATIC_RATE            (0)

#define VIADEV_DEFAULT_SRC_PATH_BITS          (0)

#define VIADEV_MAX_FAST_EAGER_SIZE            (255)

/* To handle congestion in larger clusters */
#define VIADEV_DEFAULT_RETRY_COUNT            (7)

/* Set to infinite */
#define VIADEV_DEFAULT_RNR_RETRY              (7)

#define VIADEV_DEFAULT_MAX_SG_LIST	          (20)
#define VIADEV_DEFAULT_MAX_RDMA_DST_OPS       (8)

#define HOSTNAME_LEN                          (255)

#if (defined(_MLX_PCI_EX_DDR_) || defined(_IBM_EHCA_))

    #define VIADEV_DEFAULT_MTU                (IBV_MTU_2048)

#else

    #define VIADEV_DEFAULT_MTU                (IBV_MTU_1024)

#endif

#define VIADEV_NUM_RDMA_BUFFER                (32)
#define DEFAULT_RENDEZVOUS_THRESHOLD          \
  ((VBUF_BUFFER_SIZE_DEFAULT - sizeof(viadev_packet_eager_start)))

#define RENDEZVOUS_THRESHOLD_CALC(num_proc) do {            \
    if(viadev_cluster_size==0 ||                            \
           (viadev_cluster_size==-1 && num_proc<64)) {      \
        viadev_rendezvous_threshold =                       \
           (VBUF_BUFFER_SIZE -                              \
            sizeof(viadev_packet_eager_start));             \
        viadev_num_rdma_buffer = 32;                        \
    } else if(viadev_cluster_size==1 ||                     \
           (viadev_cluster_size==-1 && num_proc<256)) {     \
        viadev_rendezvous_threshold =       (1024*16);      \
        viadev_num_rdma_buffer = 16;                        \
    } else if(viadev_cluster_size==2 ||                     \
           viadev_cluster_size==-1)    {                    \
        viadev_rendezvous_threshold =       (1024*8);       \
        viadev_num_rdma_buffer = 16;                        \
    } else {                                                \
        viadev_rendezvous_threshold =   (1024*12);          \
        viadev_num_rdma_buffer = 32;                        \
    }                                                       \
} while(0);

#define VIADEV_PREPOST_DEPTH 64
#define VIADEV_VBUF_MIN_POOL_SIZE 100
#define VIADEV_PREPOST_RENDEZVOUS_EXTRA VIADEV_PREPOST_DEPTH

#ifdef BLOCKING_SUPPORT
#define VIADEV_MAX_SPIN_COUNT                   (5000)
#endif

#ifdef MCST_SUPPORT
#define DEFAULT_MAX_CQ_SIZE (400)
#define DEFAULT_MAX_WQE    (200)
#define DEFAULT_FLOW_LABEL      (0)
#define DEFAULT_TRAFFIC_CLASS   (0)
#define DEFAULT_HOP_LIMIT       (63)
#endif

#endif                          /* _IBVERBS_CONST_H */
