/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */


/* Copyright (c) 2002-2006, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

/* Thanks to Mellanox for contributing to the Adaptive RDMA
 * implementation.
 */

#define _XOPEN_SOURCE 600

#if defined(_SMP_) || defined(USE_MPD_BASIC) || defined(USE_MPD_RING)
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#endif


#ifdef _SMP_
#ifdef _AFFINITY_
#include <sched.h>
#endif /*_AFFINITY_*/
#endif

#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include "viadev.h"
#include "viaparam.h"
#include "viapriv.h"
#include "vbuf.h"
#include "dreg.h"
#include "process/pmgr_client.h"
#include "mpid_smpi.h"

#ifdef ON_DEMAND
#include "cm.h"
#include "cm_user.h"
#endif

#ifdef MCST_SUPPORT
#include <complib/cl_debug.h>
#include "bcast_info.h"

static void ib_ud_qp_init(void);
static void ud_mcst_av_init(void);
static void alloc_ack_region(void);
static void alloc_bcast_buffers(void);
void create_mcgrp(int);
int join_mcgrp(int);
void leave_mcgrp(int);
int mlid;
void viadev_post_ud_recv(vbuf *);
void ud_cleanup(void);
int ud_prepost_count = 0;
int ud_prepost_threshold = VIADEV_UD_PREPOST_THRESHOLD;
int period = PERIOD;
int is_mcast_enabled = 1;
#endif

#ifdef USE_MPD_RING
static void ib_vapi_bootstrap_ring(int, int, char *, char *, char *,
                                   struct ibv_mr *);
static int bootstrap_barrier(void);
static void ib_vapi_enable_ring(int, int, char *);
#endif

#ifdef _SMP_
#ifdef _AFFINITY_
extern unsigned int viadev_enable_affinity;
#endif
#endif

static void ib_init(void);

#ifndef ON_DEMAND
static void ib_qp_enable(void);
static void ib_rank_lid_table(int *alladdrs, int len);
#endif

static void ib_finalize(void);

#ifdef USE_MPD_RING
static void *barrier_buff;
struct ibv_mr *barrier_mem_hndl;
#endif

#if defined(_SMP_) || defined(USE_MPD_BASIC) || defined(USE_MPD_RING)
static int get_host_id(char *myhostname, int hostname_len);
#endif

#if defined(LAZY_MEM_UNREGISTER) && !defined(MAC_OSX)
/* NOTE: Inclusion of malloc.h this makefile will cause redundant
 * declarations of malloc, free, etc.  If compiling under
 * gcc with -Wredundant-decls AND -Werror, the compilation
 * of this module will fail.  We have turned off
 * -Wredundant-decls for this module only.  If and when
 * this is fixed in glibc, the makefile should be repaired.
 */
#include <malloc.h>
#ifdef MALLOC_HOOK
/* Initialize our malloc hook */
void (*__malloc_initialize_hook) (void) = mvapich_init_malloc_hook;
#endif

#endif

#ifdef MCST_SUPPORT
bcast_info_t *bcast_info = &(viadev.bcast_info);
#endif

#ifdef ON_DEMAND

#ifdef _SMP_
int *hostnames_od;

void odu_init_SMP()
{
    int j;
    int *smpi_ptr = NULL;

#ifdef _AFFINITY_
    long N_CPUs_online;
    unsigned long affinity_mask = 1;
    unsigned long affinity_mask_len = sizeof(affinity_mask);

    /*Get number of CPU on machine */
    if ((N_CPUs_online = sysconf(_SC_NPROCESSORS_ONLN)) < 1) {
        perror("sysconf");
    }
#endif /*_AFFINITY_*/

    SET_ORIGINAL_MALLOC_HOOKS;

    smpi.local_nodes = (unsigned int *) malloc(viadev.np * sizeof(int));
    smpi_ptr = (unsigned int *) malloc(viadev.np * sizeof(int));

    if (hostnames_od == NULL || smpi.local_nodes == NULL
        || smpi_ptr == NULL) {
        error_abort_all(GEN_EXIT_ERR, "malloc: for SMP");
    }

    SAVE_MALLOC_HOOKS;
    SET_MVAPICH_MALLOC_HOOKS;

    smpi.only_one_device = 1;
    smpi.num_local_nodes = 0;
    for (j = 0; j < viadev.np; j++) {
        if (hostnames_od[viadev.me] == hostnames_od[j]) {
            if (j == viadev.me) {
                smpi.my_local_id = smpi.num_local_nodes;
#ifdef _AFFINITY_
                if (viadev_enable_affinity > 0) {
                    affinity_mask =
                        1 << (smpi.my_local_id % N_CPUs_online);
                    if (sched_setaffinity
                        (0, affinity_mask_len, &affinity_mask) < 0) {
                        perror("sched_setaffinity");
                    }
                }
#endif                          /* _AFFINITY_ */
            }
            smpi.local_nodes[j] = smpi.num_local_nodes;
            smpi_ptr[smpi.num_local_nodes] = j;
            smpi.num_local_nodes++;
        } else {
            smpi.only_one_device = 0;
            smpi.local_nodes[j] = -1;
        }
    }
    smpi.l2g_rank =
        (unsigned int *) malloc(smpi.num_local_nodes * sizeof(int));
    if (smpi.l2g_rank == NULL) {
        error_abort_all(GEN_EXIT_ERR, "odu_init_SMP");
    }
    for (j = 0; j < smpi.num_local_nodes; j++) {
        smpi.l2g_rank[j] = smpi_ptr[j];
    }
    free(smpi_ptr);
}
#endif

#endif                          /*ON_DEMAND */

int vapi_proc_info(int i, char **host, char **exename)
{
    int pid;
    pid = viadev.pids[i];
    *host = viadev.processes[i];
    *exename = viadev.execname;
    return (pid);
}

static void open_hca(void)
{
    struct ibv_device *ib_dev = NULL;
    struct ibv_device **dev_list;
    int i = 0;

    dev_list = ibv_get_device_list(NULL);

    if (!strncmp(viadev.device_name, VIADEV_INVALID_DEVICE, 32)) {

        /* User hasn't specified any HCA name
         * We will use the first available HCA */
        if (dev_list[0]) {
            ib_dev = dev_list[0];
        }
    } else {

        /* User specified a HCA, try to look for it */
        while (dev_list[i]) {

            if (!strncmp(ibv_get_device_name(dev_list[i]),
                         viadev.device_name, 32)) {
                ib_dev = dev_list[i];
                break;
            }
            i++;
        }
    }

    if (!ib_dev) {
        fprintf(stderr, "[%d] No IB devices found\n", viadev.me);
    }

    viadev.nic = ib_dev;

    viadev.context = ibv_open_device(ib_dev);

    ibv_free_device_list(dev_list);

    if (!viadev.context) {
        error_abort_all(GEN_EXIT_ERR, "Error getting HCA context\n");
    }

    viadev.ptag = ibv_alloc_pd(viadev.context);

    if (!viadev.ptag) {
        error_abort_all(GEN_EXIT_ERR, "Error getting protection domain\n");
    }
}

static void get_lid(void)
{
    struct ibv_port_attr port_attr;
    int port;

    /* Search for first active port */
    if (viadev_default_port < 0) {
        for (port = 1; port <= viadev_max_ports; port++) {
            if ((!ibv_query_port(viadev.context, port, &port_attr)) &&
                port_attr.state == IBV_PORT_ACTIVE && port_attr.lid) {
                viadev_default_port = port;
                viadev.my_hca_lid = port_attr.lid;
                return;
            }
        }
    } else {
        if ((!ibv_query_port(viadev.context, viadev_default_port,
                             &port_attr)) && (port_attr.lid)) {
            viadev.my_hca_lid = port_attr.lid;
            return;
        }
    }

    error_abort_all(GEN_EXIT_ERR,
                    "No active port found. Check your cable connection"
                    " and OpenSM status\n");
}

static struct ibv_cq *create_cq(struct ibv_comp_channel *ch)
{
    struct ibv_cq *cq_ptr;

    cq_ptr = ibv_create_cq(viadev.context, viadev_cq_size, NULL, ch, 0);

    if (!cq_ptr) {
        error_abort_all(IBV_RETURN_ERR, "Error creating CQ\n");
    }

    return cq_ptr;
}

#ifdef SRQ
static struct ibv_srq *create_srq()
{
    struct ibv_srq_init_attr srq_init_attr;
    struct ibv_srq *srq_ptr = NULL;

    memset(&srq_init_attr, 0, sizeof(srq_init_attr));

    srq_init_attr.srq_context = viadev.context;
    srq_init_attr.attr.max_wr = viadev_srq_size;
    srq_init_attr.attr.max_sge = 1;
    /* The limit value should be ignored during SRQ create */
    srq_init_attr.attr.srq_limit = viadev_srq_limit;

    srq_ptr = ibv_create_srq(viadev.ptag, &srq_init_attr);

    if (!srq_ptr) {
        error_abort_all(IBV_RETURN_ERR, "Error creating SRQ\n");
    }

    return srq_ptr;
}
#endif

int MPID_VIA_Init(int *argc, char ***argv, int *size, int *rank)
{
    int i;
#if (!(defined (ON_DEMAND)) || defined (USE_MPD_BASIC))
    int local_addr_len;
#endif
    int *alladdrs = NULL;
    int *local_addr = NULL;

#if (defined (_SMP_) && !((defined (USE_MPD_RING) || defined (USE_MPD_BASIC))))
    int hostid;
#endif

    int mypid;

#ifdef MEMORY_RELIABLE
    gen_crc_table();
    init_crc_queues();
    init_ofo_queues();
#endif

#ifdef MCST_SUPPORT
    char grp_info[GRP_INFO_LEN];

    if (NULL != getenv("DISABLE_HARDWARE_MCST")) {
        /* The user has disabled Multicast */
        is_mcast_enabled = 0;
    } else {
        /* The user has enabled Multicast */
        is_mcast_enabled = 1;
    }
#endif

#if defined(VIADEV_RPUT_SUPPORT) || defined(VIADEV_RGET_SUPPORT)
#ifdef LAZY_MEM_UNREGISTER
    /* Set glibc/stdlib malloc options to prevent handing
     * memory back to the system (brk) upon free.
     * Also, dont allow MMAP memory for large allocations.
     * This because we dont unregister memory when refcounts
     * go to zero.
     * Only need these options when RDMA operations are being
     * used.  Otherwise, MVICH layer owns all registered memory
     * and we will only free it if it is de-registered anyway.
     */
#if defined(MALLOC_HOOK)
    create_hash_table();
#else
    mallopt(M_TRIM_THRESHOLD, -1);
    mallopt(M_MMAP_MAX, 0);
#endif                          /* MALLOC_HOOK */

#endif                          /* LAZY_MEM_UNREGISTER */
#endif                          /* VIADEV_RPUT_SUPPORT || VIADEV_RGET_SUPPORT */

    /* init the VIA barrier counter */
    viadev.barrier_id = 0;

    /* Device is not yet initialized */
    viadev.initialized = 0;

    /*
     * Find out from the process manager who we are. 
     */

    if (!pmgr_client_init(argc, argv, &viadev.np, &viadev.me,
                          &viadev.global_id, &viadev.processes)) {
        error_abort_all(GEN_EXIT_ERR, "pmgr_client_init failed");
    }

    *size = viadev.np;
    *rank = viadev.me;
    strcpy(viadev.execname, ((const char **) (*argv))[0]);
    viadev.is_finalized = 0;

#if (defined(SRQ) || defined(ON_DEMAND))
    init_vbuf_lock();
#endif

    /* Now that we know our rank, we should also know
     * our name */
    viadev.my_name = (char *) malloc(sizeof(char) * HOSTNAME_LEN);

    if (NULL == viadev.my_name) {
        error_abort_all(GEN_EXIT_ERR, "malloc failed");
    }

    if (gethostname(viadev.my_name, sizeof(char) * HOSTNAME_LEN) < 0) {
        error_abort_all(GEN_EXIT_ERR, "gethostname failed");
    }

    D_PRINT("size=%d, myrank=%d\n", *size, *rank);


    SET_ORIGINAL_MALLOC_HOOKS;

    /* For both SMP and no-SMP, we have the same address length */
    alladdrs = (int *) malloc(3 * viadev.np * sizeof(int));
    local_addr = (int *) malloc(2 * viadev.np * sizeof(int));

    viadev.lid_table = (uint16_t *) malloc(viadev.np * sizeof(int));
    viadev.qp_table = (uint32_t *) malloc(viadev.np * sizeof(int));
    viadev.qp_hndl =
        (struct ibv_qp **) malloc(viadev.np * sizeof(struct ibv_qp *));
    viadev.pids = (int *) malloc(viadev.np * sizeof(int));
    viadev.pids[viadev.me] = mypid = (int) getpid();

    if (alladdrs == NULL || local_addr == NULL ||
        viadev.lid_table == NULL || viadev.qp_table == NULL) {
        error_abort_all(GEN_EXIT_ERR,
                        "malloc for alladdrs/local_addr/lid_table/qp_table");
    }

    SAVE_MALLOC_HOOKS;
    SET_MVAPICH_MALLOC_HOOKS;

    /* set run-time parameters that have been passed in through
     * environment variables */
    viadev_init_parameters(viadev.np, viadev.me);

    viadev.maxtransfersize = viadev_max_rdma_size;

    open_hca();
    get_lid();

    /* Since we need to excahge qp # also, 
     * we will create QP first, exchange addresses and 
     * then enable them */

    ib_init();

    viadev.array_recv_desc = (struct ibv_recv_wr *)
        malloc(sizeof(struct ibv_recv_wr) * viadev_rq_size);

    memset(viadev.array_recv_desc, '0',
           sizeof(struct ibv_recv_wr) * viadev_rq_size);

#ifdef ON_DEMAND
    {
        MPICM_ib_context ib_context;
        struct ibv_qp_attr attr;
        struct ibv_qp_init_attr init_attr;

        viadev.ud_qpn_table = malloc(sizeof(uint32_t) * viadev.np);
        viadev.conn_table = malloc(sizeof(MPICM_conn_type) * viadev.np);
        viadev.pending_req_head =
            malloc(sizeof(cm_pending_request *) * viadev.np);
        viadev.pending_req_tail =
            malloc(sizeof(cm_pending_request *) * viadev.np);

        for (i = 0; i < viadev.np; i++) {
            viadev.conn_table[i] = MPICM_IB_NONE;
            viadev.pending_req_head[i] = NULL;
            viadev.pending_req_tail[i] = NULL;
        }

        ib_context.context = viadev.context;
        ib_context.hca_port = viadev_default_port;
        ib_context.pd = viadev.ptag;
        ib_context.size = viadev.np;
        ib_context.rank = viadev.me;
        ib_context.confim_recv_wr_id = MAGIC_WRID_RECV;
        ib_context.confim_send_wr_id = MAGIC_WRID_SEND;

        /*Setup qp attributes */
        /*Create */
        memset(&init_attr, 0, sizeof(struct ibv_qp_init_attr));
        init_attr.send_cq = viadev.cq_hndl;
        init_attr.recv_cq = viadev.cq_hndl;
#ifdef SRQ
        init_attr.srq = viadev.srq_hndl;
        init_attr.cap.max_recv_wr = 0;
#else
        init_attr.srq = NULL;
        init_attr.cap.max_recv_wr = viadev_rq_size;
#endif
        init_attr.cap.max_send_wr = viadev_sq_size;
        init_attr.cap.max_send_sge = 20;
        init_attr.cap.max_recv_sge = 1;
        init_attr.cap.max_inline_data = viadev_max_inline_size;
        init_attr.qp_type = IBV_QPT_RC;
        ib_context.rc_qp_init_attr = init_attr;


        /*Init */
        memset(&attr, 0, sizeof(struct ibv_qp_attr));
        attr.qp_state = IBV_QPS_INIT;
        attr.pkey_index = 0;
        attr.port_num = viadev_default_port;
        attr.qp_access_flags = IBV_ACCESS_REMOTE_WRITE |
            IBV_ACCESS_REMOTE_READ;
        ib_context.rc_qp_attr_to_init = attr;
        ib_context.rc_qp_mask_to_init = IBV_QP_STATE |
            IBV_QP_PKEY_INDEX | IBV_QP_PORT | IBV_QP_ACCESS_FLAGS;

         /*RTR*/ memset(&attr, 0, sizeof(struct ibv_qp_attr));
        attr.qp_state = IBV_QPS_RTR;
        attr.path_mtu = viadev_default_mtu;
        attr.max_dest_rd_atomic = viadev_default_qp_ous_rd_atom;
        attr.min_rnr_timer = viadev_default_min_rnr_timer;
        attr.ah_attr.sl = viadev_default_service_level;
        attr.ah_attr.src_path_bits = viadev_default_src_path_bits;
        attr.ah_attr.port_num = viadev_default_port;
        attr.ah_attr.static_rate = viadev_default_static_rate;

        ib_context.rc_qp_attr_to_rtr = attr;
        ib_context.rc_qp_mask_to_rtr = IBV_QP_STATE |
            IBV_QP_AV |
            IBV_QP_PATH_MTU |
            IBV_QP_DEST_QPN |
            IBV_QP_RQ_PSN |
            IBV_QP_MAX_DEST_RD_ATOMIC | IBV_QP_MIN_RNR_TIMER;


         /*RTS*/ memset(&attr, 0, sizeof(struct ibv_qp_attr));
        attr.qp_state = IBV_QPS_RTS;
        attr.timeout = viadev_default_time_out;
        attr.retry_cnt = viadev_default_retry_count;
        attr.rnr_retry = viadev_default_rnr_retry;
        attr.sq_psn = viadev_default_psn;
        attr.max_rd_atomic = viadev_default_qp_ous_rd_atom;

        ib_context.rc_qp_attr_to_rts = attr;
        ib_context.rc_qp_mask_to_rts = IBV_QP_STATE |
            IBV_QP_TIMEOUT |
            IBV_QP_RETRY_CNT |
            IBV_QP_RNR_RETRY | IBV_QP_SQ_PSN | IBV_QP_MAX_QP_RD_ATOMIC;

        /*Initialize UD-based CM */
        if (MPICM_Init_UD(&ib_context, &(viadev.ud_qpn_table[viadev.me]))) {
            error_abort_all(GEN_EXIT_ERR, "MPICM_Init_UD");
        }
    }
#endif

#ifdef MCST_SUPPORT

    if (is_mcast_enabled) {

        viadev.mcg_dgid.raw[0] = 0xFF;
        viadev.mcg_dgid.raw[1] = 0x00;
        viadev.mcg_dgid.raw[2] = 0x00;
        viadev.mcg_dgid.raw[3] = 0x00;
        viadev.mcg_dgid.raw[4] = 0x00;
        viadev.mcg_dgid.raw[5] = 0x00;
        viadev.mcg_dgid.raw[6] = 0x00;
        viadev.mcg_dgid.raw[7] = 0x00;
        viadev.mcg_dgid.raw[8] = 0x00;
        viadev.mcg_dgid.raw[9] = 0x00;
        viadev.mcg_dgid.raw[10] = 0x00;
        viadev.mcg_dgid.raw[11] = 0x00;
        viadev.mcg_dgid.raw[12] = 0x00;
        viadev.mcg_dgid.raw[13] = 0x00;
        viadev.mcg_dgid.raw[14] = 0x00;
        viadev.mcg_dgid.raw[15] = 0x00;


        ib_ud_qp_init();        /* Allocate UD-QP */

#if defined(MINISM)
        ud_mcst_av_init();      /* Allocate address vect.for  multicast */
#endif
        alloc_ack_region();     /* Allocate Ack region for reliability  */

        alloc_bcast_buffers();  /* Allocate Bcast,Reliability structures  */
    }
#endif

#ifdef ON_DEMAND
#ifdef _SMP_
    hostnames_od = (int *) malloc(viadev.np * sizeof(int));
    if (hostnames_od == NULL) {
        error_abort_all(GEN_EXIT_ERR, "malloc: hostnames_od for SMP");
    }
#endif
#if defined(USE_MPD_RING) || defined (USE_MPD_BASIC)
#if defined(USE_MPD_BASIC)
    {
        int temp_hostid;
        char *alladdrs_self;
        char *alladdrs_group;
        char *temp_ptr;
#ifdef _SMP_
        local_addr_len = 3 * 8;
#else
        local_addr_len = 2 * 8;
#endif
        alladdrs_self = (char *) malloc(local_addr_len + 4);
        alladdrs_group = (char *) malloc(viadev.np * local_addr_len + 4);

        /* Prepare all the queue pairs from myself */
        temp_ptr = (char *) alladdrs_self;
        sprintf(temp_ptr, "%08d", viadev.ud_qpn_table[viadev.me]);
        temp_ptr += 8;
        sprintf(temp_ptr, "%08d", viadev.my_hca_lid);
        temp_ptr += 8;
#ifdef _SMP_
        temp_hostid = get_host_id(viadev.my_name, HOSTNAME_LEN);
        sprintf(temp_ptr, "%08X", temp_hostid);
        temp_ptr += 8;
#endif
        *temp_ptr = '\0';
        pmgr_exchange_addresses(alladdrs_self, local_addr_len,
                                alladdrs_group, viadev.pids);

        for (i = 0; i < viadev.np; i++) {
            char temp_buf[9];
            char *temp = alladdrs_group + local_addr_len * i;
            strncpy(temp_buf, temp, 8);
            temp_buf[8] = '\0';
            viadev.ud_qpn_table[i] = atoi(temp_buf);
            strncpy(temp_buf, temp + 8, 8);
            temp_buf[8] = '\0';
            viadev.lid_table[i] = atoi(temp_buf);
#ifdef _SMP_
            strncpy(temp_buf, temp + 16, 8);
            temp_buf[8] = '\0';
            hostnames_od[i] = atoi(temp_buf);
#endif
        }

        free(alladdrs_self);
        free(alladdrs_group);
    }

#else
        /* NO MPD_RING for ON_DEMAND*/
        error_abort_all(GEN_EXIT_ERR, "MPD_RING setup is not valid with ON_DEMAND");
#endif                          /*USE_MPD_BASIC */

#else                           /*No MPD */
    {
        int addr_len = (viadev.np + 1) * sizeof(uint32_t);
        uint32_t *addr_self =
            (uint32_t *) malloc(addr_len + sizeof(uint32_t));
        uint32_t *addr_all = (uint32_t *) malloc(3 * addr_len);

        memset(addr_self, 0, addr_len + sizeof(uint32_t));
        memset(addr_all, 0, 3 * addr_len);

        for (i = 0; i < viadev.np; i++) {
            if (i == viadev.me)
                addr_self[i] = viadev.my_hca_lid;
            else                /*Same for every one */
                addr_self[i] = viadev.ud_qpn_table[viadev.me];
        }
#ifdef _SMP_
        hostid = get_host_id(viadev.my_name, HOSTNAME_LEN);
        addr_self[viadev.np] = hostid;
#else                           /*dummy integer */
        addr_self[viadev.np] = viadev.me;
#endif
        pmgr_exchange_addresses(addr_self, addr_len, addr_all,
                                viadev.pids);

        for (i = 0; i < viadev.np; i++) {
            viadev.lid_table[i] = addr_all[i];
            if (i != viadev.me)
                viadev.ud_qpn_table[i] = addr_all[viadev.np + i];
#ifdef _SMP_
            hostnames_od[i] = addr_all[viadev.np * 2 + i];
#endif
        }
        free(addr_self);
        free(addr_all);
    }
#endif /*MPD*/
        if (MPICM_Connect_UD(viadev.ud_qpn_table, viadev.lid_table)) {
        error_abort_all(GEN_EXIT_ERR, "MPICM_Connect_UD");
    }

    {                          
        /* ALLOCATING SEND/RECV VBUFS */
        allocate_vbufs(viadev_vbuf_pool_size);

        SET_ORIGINAL_MALLOC_HOOKS;

        /* initialize connection parameters */
        viadev.connections = (viadev_connection_t *)
            malloc(sizeof(viadev_connection_t) * viadev.np);

        if (NULL == viadev.connections) {
            error_abort_all(GEN_EXIT_ERR,
                            "Unable to malloc connection structures");
        }

        memset(viadev.connections, 0,
               (sizeof(viadev_connection_t) * viadev.np));

        /* Initialize the NULL values for connection to itself*/
        /* these should be ignored for self-communication */
        viadev.connections[viadev.me].vi = NULL;
        viadev.connections[viadev.me].remote_credit = 0;
        viadev.connections[viadev.me].remote_cc = 0;
        viadev.connections[viadev.me].local_credit = 0;
        
#if defined(ADAPTIVE_RDMA_FAST_PATH)
        viadev.RDMA_polling_group_size = 0;
        viadev.RDMA_polling_group =
            (int *) malloc(viadev.np * sizeof(int));

        if (NULL == viadev.RDMA_polling_group) {
            error_abort_all(GEN_EXIT_ERR,
                            "Unable to malloc connection structures");
        }
#endif

        SAVE_MALLOC_HOOKS;
        SET_MVAPICH_MALLOC_HOOKS;

#ifdef SRQ
        pthread_spin_init(&viadev.srq_post_lock, 0);
        pthread_spin_lock(&viadev.srq_post_lock);

        viadev.posted_bufs = viadev_post_srq_buffers(viadev_srq_size);

        {
            struct ibv_srq_attr srq_attr;
            srq_attr.max_wr = viadev_srq_size;
            srq_attr.max_sge = 1;
            srq_attr.srq_limit = viadev_srq_limit;

            if (ibv_modify_srq(viadev.srq_hndl, &srq_attr, IBV_SRQ_LIMIT)) {
                error_abort_all(IBV_RETURN_ERR,
                                "Couldn't modify SRQ limit\n");
            }

            /* Start the async thread which watches for SRQ limit events */
            pthread_create(&viadev.async_thread, NULL,
                           (void *) async_thread, (void *) viadev.context);
        }

        pthread_spin_unlock(&viadev.srq_post_lock);
#endif
    }
    if (MPICM_Listen(MPICM_IB_RC_PT2PT, cm_server_callback, NULL)) {
        error_abort_all(GEN_EXIT_ERR, "MPICM_Listen");
    }
    pmgr_barrier();

#ifdef _SMP_
    odu_init_SMP();
    free(hostnames_od);
#endif

#else                           /*No  ON_DEMAND */

#if defined(USE_MPD_RING) || defined (USE_MPD_BASIC)
    /* 
     * For the scalable network discovery, we create a ring of queue pairs.
     * Then we exchange other queue_pair #'s over the ring. 
     * In future, consider using MPD as a run_time option later. 
     * For now, we stick with some of the process manager code.
     */

    {

        int temp_hostid;
        char *alladdrs_self;
        char *alladdrs_group;
        char *temp_ptr;
        struct ibv_mr *alladdrs_mr_hndl;


#ifdef USE_MPD_RING
        int lhs, rhs;
        char ring_qp_out[64];
        char ring_qp_in[128];

        temp_ptr = (char *) ring_qp_out;
        lhs = (viadev.me + viadev.np - 1) % viadev.np;
        rhs = (viadev.me + 1) % viadev.np;

        /* Prepare my_hca_lid, lhs qp, rsh qp for exchange */
        sprintf(temp_ptr, "%08d", viadev.my_hca_lid);
        temp_ptr += 8;
        sprintf(temp_ptr, "%08d", viadev.boot_qp_hndl[0]->qp_num);
        temp_ptr += 8;
        sprintf(temp_ptr, "%08d", viadev.boot_qp_hndl[1]->qp_num);
        temp_ptr += 8;
        *temp_ptr = '\0';
        D_PRINT("ring_qp_out has %s\n", ring_qp_out);
#endif

        local_addr_len = (viadev.np + 1) * QPLEN_XDR;
        alladdrs_self = (char *) malloc(local_addr_len + 4);
        alladdrs_group = (char *) malloc(viadev.np * local_addr_len + 4);

        alladdrs_mr_hndl = ibv_reg_mr(viadev.ptag, alladdrs_group,
                                      viadev.np * local_addr_len,
                                      IBV_ACCESS_LOCAL_WRITE |
                                      IBV_ACCESS_REMOTE_WRITE);

        if (!alladdrs_mr_hndl) {
            error_abort_all(IBV_RETURN_ERR, "Couldn't register memory\n");
        }

        /* Prepare all the queue pairs from myself */
        temp_ptr = (char *) alladdrs_self;
        for (i = 0; i < viadev.np; i++) {

            /* Queue pairs for all processes */
            if (i == viadev.me) {
                sprintf(temp_ptr, "%08d", viadev.my_hca_lid);
            } else {
                sprintf(temp_ptr, "%08d", viadev.qp_hndl[i]->qp_num);
            }
            temp_ptr += 8;
        }
        temp_hostid = get_host_id(viadev.my_name, HOSTNAME_LEN);
        sprintf(temp_ptr, "%08X", temp_hostid);
        temp_ptr += 8;
        *temp_ptr = '\0';


#ifdef USE_MPD_BASIC
        {
            int self_length = strlen(alladdrs_self);
            assert(self_length == local_addr_len);
            pmgr_exchange_addresses(alladdrs_self, self_length,
                                    alladdrs_group, viadev.pids);
        }
#else
        D_PRINT("Before exchange, out=%s\n", ring_qp_out);
        /* Save the static buffer address */
        barrier_buff = (void *) alladdrs_group;
        barrier_mem_hndl = alladdrs_mr_hndl;

        pmgr_exchange_addresses(ring_qp_out, strlen(ring_qp_out),
                                ring_qp_in, viadev.pids);
        D_PRINT("After exchange, in=%s\n", ring_qp_in);

        /* Enable the ring to exhange all the others queue_pair numbers.
         * Consider rewriting all the following functions.
         */
        ib_vapi_bootstrap_ring(lhs, rhs, ring_qp_in, alladdrs_self,
                               alladdrs_group, alladdrs_mr_hndl);
#endif
        D_PRINT("Done exchanging addresses group=%s\n", alladdrs_group);

        /*The memory allocated for alladdrs is not freed at this time. */
        ib_rank_lid_table((int *) alladdrs_group, local_addr_len);
        ib_qp_enable();

        /* Free all the temorary memory used for MPD */
        ibv_dereg_mr(alladdrs_mr_hndl);

        free(alladdrs_self);
        free(alladdrs_group);
    }
#else                           /* (defined(USE_MPD_BASIC) || defined (USE_MPD_RING)) */

    for (i = 0; i < viadev.np; ++i) {

        if (i == viadev.me) {
            local_addr[i] = viadev.my_hca_lid;
        } else {
            local_addr[i] = viadev.qp_hndl[i]->qp_num;
        }
    }

    local_addr_len = viadev.np * sizeof(int);

#ifdef _SMP_
    hostid = get_host_id(viadev.my_name, HOSTNAME_LEN);
    local_addr[viadev.np] = hostid;
    local_addr_len += sizeof(int);

#else                           /* put a useless integer for no-SMP mode */

    local_addr[viadev.np] = viadev.me;
    local_addr_len += sizeof(int);
#endif

    pmgr_exchange_addresses(local_addr, local_addr_len, alladdrs,
                            viadev.pids);

    ib_rank_lid_table(alladdrs, local_addr_len);
    ib_qp_enable();

#endif                          /* End of ! (defined(USE_MPD_BASIC) || defined (USE_MPD_RING)) */


    free(local_addr);
    free(alladdrs);

#endif                          /*ON_DEMAND */


#if (defined(VIADEV_RPUT_SUPPORT) || defined(VIADEV_RGET_SUPPORT))
    dreg_init();
#endif

#if defined(MCST_SUPPORT)

    if (is_mcast_enabled) {
        if (viadev.me == 0) {
            create_mcgrp(viadev_default_port);  /* Creation of mcst group */
        }

        if (viadev.me == 0) {
            strncpy(grp_info, "MCS_Grp1", GRP_INFO_LEN);
        }
    }

    pmgr_exchange_mcs_group_sync(grp_info, GRP_INFO_LEN, 0);

    if (is_mcast_enabled) {

        mlid = join_mcgrp(viadev_default_port); /* Joining the group */

        ud_mcst_av_init();      /* Allocate address vect. for multicast */

        /* Preposting the descriptors for the UD transport */
        for (i = 0; i < VIADEV_UD_PREPOST_DEPTH; i++) {
            PREPOST_UD_VBUF_RECV();
            ud_prepost_count++;
        }
    }
#endif                          /*MCST_SUPPORT */

    viadev.coll_comm_reg = NULL;

    return 0;
}

/*
 * Set up all communication and put VIs in the VI array in viadev
 * The algorithm is as follows (may change when M-VIA supports peer-to-peer
 * connections):
 *   . use client-server routines to establish connections
 *   . for each pair, the lower-numbered process acts as the server
 *   . we should be careful not to establish hotspots. E.g., if
 *     every process connects with others in order 0, 1, 2. We
 *     should also be careful not to serialize. 
 *     there is a nice fancy way to do this, with a communication pattern
 *     essentially the same as in an FFT. 
 *   . However, it is a bit tricky to get right so for now 
 *     do the stupid serialized version (xxx):
 *        for (i = 0; i < me; i++) connect(i)
 *        for (i = me+1; i <= np; i++) accept(i)
 *   
 *  Note that we do not set up a connection to ourselves 
 *
 */

void MPID_Wtime(double *t)
{
    struct timeval tv;
    static int initialized = 0;
    static int sec_base;
    gettimeofday(&tv, NULL);
    if (!initialized) {
        sec_base = tv.tv_sec;
        initialized = 1;
    }
    *t = (double) (tv.tv_sec - sec_base) + (double) tv.tv_usec * 1.0e-6;
}

void MPID_Wtick(double *t)
{
    *t = VIADEV_WTICK;
}

void MPID_VIA_Finalize(void)
{
    /* Unpinning any leftover collective buffers */
    reg_entry *current = viadev.coll_comm_reg;
    reg_entry *tmp;

    while (current != NULL) {
        tmp = current;
        if (current->valid == 1) {
            deregister_memory(current->mem_hndl);
        }
        current = current->next;
        free(tmp);
    }


#ifdef BLOCKING_SUPPORT

    if (ibv_destroy_comp_channel(viadev.comp_channel)) {
        error_abort_all(IBV_RETURN_ERR, "Cannot clear completion handler");
    }
#endif

    /* VAPI finalize called */
    ib_finalize();

    /* free connection array space */
    if (viadev.connections) {
        free(viadev.connections);
    }

    /* let the process manager reclaim data structures 
     * before termination */

    pmgr_finalize();

    if (viadev.lid_table) {
        free(viadev.lid_table);
    }
    if (viadev.qp_table) {
        free(viadev.qp_table);
    }
    if (viadev.pids) {
        free(viadev.pids);
    }
    if (viadev.processes) {
        free(viadev.processes);
    }
#ifdef _SMP_
    if (smpi.local_nodes) {
        free(smpi.local_nodes);
    }
    if (smpi.l2g_rank) {
        free(smpi.l2g_rank);
    }
#endif

#ifdef MCST_SUPPORT
    /* complib_exit() added for fixing comlib issue on RH-AS3,RH90,SLES9.
       Works only with osm-mvapi-0.3.2-rc1 */
    complib_exit();
#endif


    D_PRINT("exiting MPID_VIA_Finalize...\n");

}

/* internal functions used for initialization 
 * and finalization by VAPI */
static void ib_init(void)
{
#ifndef ON_DEMAND
    int i;
    struct ibv_qp_attr qp_attr;
    struct ibv_qp_init_attr attr;
#endif

#ifdef BLOCKING_SUPPORT
    viadev.comp_channel = ibv_create_comp_channel(viadev.context);

    if (!viadev.comp_channel) {
        error_abort_all(IBV_RETURN_ERR,
                        "Cannot create completion channel\n");
    }

    viadev.cq_hndl = create_cq(viadev.comp_channel);

    if (ibv_req_notify_cq(viadev.cq_hndl, 0)) {
        error_abort_all(IBV_RETURN_ERR,
                        "Cannot request completion notification\n");
    }
#else
    viadev.cq_hndl = create_cq(NULL);
#endif


#ifdef USE_MPD_RING
    viadev.boot_cq_hndl = create_cq(NULL);
#endif

#ifdef SRQ
    viadev.srq_hndl = create_srq();
#endif

#ifndef ON_DEMAND
    for (i = 0; i < viadev.np; i++) {

        memset(&attr, 0, sizeof(attr));
        attr.send_cq = viadev.cq_hndl;
        attr.recv_cq = viadev.cq_hndl;
#ifdef SRQ
        attr.srq = viadev.srq_hndl;
        attr.cap.max_recv_wr = 0;
#else
        attr.cap.max_recv_wr = viadev_rq_size;
#endif
        attr.cap.max_send_wr = viadev_sq_size;
        attr.cap.max_send_sge = 20;
        attr.cap.max_recv_sge = 1;
#ifndef _PPC64_
        attr.cap.max_inline_data = viadev_max_inline_size;
#else
        attr.cap.max_inline_data = -1;
#endif
        attr.qp_type = IBV_QPT_RC;

        if (i == viadev.me) {
            continue;
        }

        viadev.qp_hndl[i] = ibv_create_qp(viadev.ptag, &attr);

        if (!viadev.qp_hndl[i]) {
            error_abort_all(GEN_EXIT_ERR, "Error creating QP\n");
        }
    }

    for (i = 0; i < viadev.np; i++) {

        if (i == viadev.me) {
            continue;
        }

        qp_attr.qp_state = IBV_QPS_INIT;
        qp_attr.pkey_index = 0;
        qp_attr.port_num = viadev_default_port;
        qp_attr.qp_access_flags = IBV_ACCESS_REMOTE_WRITE |
            IBV_ACCESS_REMOTE_READ;

        if (ibv_modify_qp(viadev.qp_hndl[i], &qp_attr,
                          IBV_QP_STATE |
                          IBV_QP_PKEY_INDEX |
                          IBV_QP_PORT | IBV_QP_ACCESS_FLAGS)) {
            error_abort_all(GEN_EXIT_ERR, "Couldn't modify QP to init\n");
        }
    }

#endif

#ifdef USE_MPD_RING
    memset(&attr, 0, sizeof(attr));
    attr.qp_type = IBV_QPT_RC;
    attr.sq_sig_all = 0;
    attr.cap.max_send_wr = viadev_sq_size;
    attr.cap.max_recv_wr = viadev_rq_size;
    attr.cap.max_send_sge = viadev_default_max_sg_list;
    attr.cap.max_recv_sge = viadev_default_max_sg_list;
#ifndef _PPC64_
    attr.cap.max_inline_data = viadev_max_inline_size;
#else
    attr.cap.max_inline_data = -1;
#endif
    attr.send_cq = viadev.boot_cq_hndl;
    attr.recv_cq = viadev.boot_cq_hndl;

    viadev.boot_qp_hndl[0] = ibv_create_qp(viadev.ptag, &attr);

    if (!viadev.boot_qp_hndl[0]) {
        error_abort_all(GEN_EXIT_ERR, "Error creating QP\n");
    }

    viadev.boot_qp_hndl[1] = ibv_create_qp(viadev.ptag, &attr);

    if (!viadev.boot_qp_hndl[1]) {
        error_abort_all(GEN_EXIT_ERR, "Error creating QP\n");
    }
#endif
}

#ifndef ON_DEMAND
static void ib_qp_enable(void)
{
    int i;
    struct ibv_qp_attr attr;
#ifdef RDMA_FAST_PATH
    vbuf *v;
    int k;
#endif
#ifndef SRQ
    int j;
#endif

#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
    int pagesize;
    int vbuf_alignment;
#endif

    for (i = 0; i < viadev.np; i++) {

        memset(&attr, 0, sizeof(attr));

        attr.qp_state = IBV_QPS_RTR;
        attr.path_mtu = viadev_default_mtu;
        attr.max_dest_rd_atomic = viadev_default_qp_ous_rd_atom;
        attr.min_rnr_timer = viadev_default_min_rnr_timer;
        attr.ah_attr.sl = viadev_default_service_level;
        attr.ah_attr.src_path_bits = viadev_default_src_path_bits;
        attr.ah_attr.port_num = viadev_default_port;
        attr.ah_attr.static_rate = viadev_default_static_rate;

        if (i == viadev.me) {
            continue;
        }

        attr.dest_qp_num = viadev.qp_table[i];
        attr.ah_attr.dlid = viadev.lid_table[i];

        if (ibv_modify_qp(viadev.qp_hndl[i], &attr,
                          IBV_QP_STATE |
                          IBV_QP_AV |
                          IBV_QP_PATH_MTU |
                          IBV_QP_DEST_QPN |
                          IBV_QP_RQ_PSN |
                          IBV_QP_MAX_DEST_RD_ATOMIC |
                          IBV_QP_MIN_RNR_TIMER)) {
            error_abort_all(GEN_EXIT_ERR, "Failed to modify QP to RTR\n");
        }
    }

    for (i = 0; i < viadev.np; i++) {

        memset(&attr, 0, sizeof(attr));

        attr.qp_state = IBV_QPS_RTS;
        attr.timeout = viadev_default_time_out;
        attr.retry_cnt = viadev_default_retry_count;
        attr.rnr_retry = viadev_default_rnr_retry;
        attr.sq_psn = viadev_default_psn;
        attr.max_rd_atomic = viadev_default_qp_ous_rd_atom;

        if (i == viadev.me) {
            continue;
        }

        if (ibv_modify_qp(viadev.qp_hndl[i], &attr,
                          IBV_QP_STATE |
                          IBV_QP_TIMEOUT |
                          IBV_QP_RETRY_CNT |
                          IBV_QP_RNR_RETRY |
                          IBV_QP_SQ_PSN | IBV_QP_MAX_QP_RD_ATOMIC)) {
            error_abort_all(GEN_EXIT_ERR, "Failed to modify QP to RTS\n");
        }
    }


    /****************prepost receive descriptors ****************/

    /* ALLOCATING SEND/RECV VBUFS */
    allocate_vbufs(viadev_vbuf_pool_size);

    SET_ORIGINAL_MALLOC_HOOKS;

    /* initialize connection parameters */
    viadev.connections = (viadev_connection_t *)
        malloc(sizeof(viadev_connection_t) * viadev.np);

    if (NULL == viadev.connections) {
        error_abort_all(GEN_EXIT_ERR,
                        "Unable to malloc connection structures");
    }

    memset(viadev.connections, 0,
           (sizeof(viadev_connection_t) * viadev.np));

#if defined(ADAPTIVE_RDMA_FAST_PATH)
    viadev.RDMA_polling_group_size = 0;
    viadev.RDMA_polling_group = (int *) malloc(viadev.np * sizeof(int));

    if (NULL == viadev.RDMA_polling_group) {
        error_abort_all(GEN_EXIT_ERR,
                        "Unable to malloc connection structures");
    }
#endif

    SAVE_MALLOC_HOOKS;
    SET_MVAPICH_MALLOC_HOOKS;

#ifdef SRQ
    pthread_spin_init(&viadev.srq_post_lock, 0);
    pthread_spin_lock(&viadev.srq_post_lock);

    viadev.posted_bufs = viadev_post_srq_buffers(viadev_srq_size);

    {
        struct ibv_srq_attr srq_attr;
        srq_attr.max_wr = viadev_srq_size;
        srq_attr.max_sge = 1;
        srq_attr.srq_limit = viadev_srq_limit;

        if (ibv_modify_srq(viadev.srq_hndl, &srq_attr, IBV_SRQ_LIMIT)) {
            error_abort_all(IBV_RETURN_ERR, "Couldn't modify SRQ limit\n");
        }

        /* Start the async thread which watches for SRQ limit events */
        pthread_create(&viadev.async_thread, NULL,
                       (void *) async_thread, (void *) viadev.context);
    }

    pthread_spin_unlock(&viadev.srq_post_lock);

#endif

    for (i = 0; i < viadev.np; i++) {

        viadev_connection_t *c = &viadev.connections[i];
        c->global_rank = i;
        c->next_packet_expected = 1;
        c->next_packet_tosend = 1;
        c->shandle_head = NULL;
        c->shandle_tail = NULL;
        c->rhandle = NULL;
        c->nextflow = NULL;
        c->inflow = 0;
        c->preposts = 0;
        c->send_wqes_avail = viadev_sq_size;
        c->ext_sendq_head = NULL;
        c->ext_sendq_tail = NULL;
        c->rendezvous_packets_expected = 0;
        viadev_init_backlog_queue(c);

#if defined(VIADEV_RGET_SUPPORT)
        c->rdma_reads_avail = viadev_default_qp_ous_rd_atom;
        c->ext_rdma_read_head = NULL;
        c->ext_rdma_read_tail = NULL;
#endif

#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
#ifndef DISABLE_HEADER_CACHING
        /* init cached packet header */
        memset(&(c->cached_outgoing), 0,
               sizeof(viadev_packet_eager_start));
        memset(&(c->cached_incoming), 0,
               sizeof(viadev_packet_eager_start));
        c->cached_hit = c->cached_miss = 0;
        c->cached_incoming.header.type = c->cached_outgoing.header.type
            = FAST_EAGER_CACHED;
#endif
#endif

        if (i != viadev.me) {
            c->vi = viadev.qp_hndl[i];

#ifndef SRQ
            /* prepost receives */
            for (j = 0; j < viadev_initial_prepost_depth; j++) {
                PREPOST_VBUF_RECV(c);
            }
#endif

            c->remote_credit = viadev_initial_credits;
            c->remote_cc = viadev_initial_credits;
            c->local_credit = 0;
            c->preposts = viadev_initial_prepost_depth;
#ifdef SRQ
            c->initialized = 1;
            c->pending_r3_sends = 0;
#else
            c->initialized = (viadev_initial_prepost_depth
                              ==
                              viadev_prepost_depth +
                              viadev_prepost_noop_extra);
#endif
        } else {
            /* these should be ignored for self-communication */
            c->vi = NULL;
            c->remote_credit = 0;
            c->remote_cc = 0;
            c->local_credit = 0;
        }

#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
        vbuf_alignment = 64;
        pagesize = getpagesize();

        if (i != viadev.me) {

            /* initialize revelant fields */
            c->rdma_credit = 0;
            c->num_no_completion = 0;

            if (viadev_num_rdma_buffer) {

#ifdef ADAPTIVE_RDMA_FAST_PATH
                /* By default, all pointers are cleared, indicating no buffers
                 * at the receive and no credits at the sender side. Only if 
                 * RDMA eager buffers are allocated, this pointers will be updated
                 */
                c->phead_RDMA_send = 0;
                c->ptail_RDMA_send = 0;
                c->p_RDMA_recv = 0;
                c->p_RDMA_recv_tail = 0;

#else

                SET_ORIGINAL_MALLOC_HOOKS;

                /* allocate vbuf struct buffers */
                if (posix_memalign
                    ((void **) &c->RDMA_send_buf, vbuf_alignment,
                     sizeof(struct vbuf) * viadev_num_rdma_buffer)) {
                    error_abort_all(GEN_EXIT_ERR,
                                    "Unable to malloc vbuf structures");
                }

                if (posix_memalign
                    ((void **) &c->RDMA_recv_buf, vbuf_alignment,
                     sizeof(struct vbuf) * viadev_num_rdma_buffer)) {
                    error_abort_all(GEN_EXIT_ERR,
                                    "Unable to malloc vbuf structures");
                }

                /* zero buffers */
                memset(c->RDMA_send_buf, 0,
                       sizeof(struct vbuf) * viadev_num_rdma_buffer);
                memset(c->RDMA_recv_buf, 0,
                       sizeof(struct vbuf) * viadev_num_rdma_buffer);

                if (posix_memalign
                    ((void **) &c->RDMA_send_buf_DMA, pagesize,
                     viadev_vbuf_total_size * viadev_num_rdma_buffer)) {
                    error_abort_all(GEN_EXIT_ERR,
                                    "Unable to malloc rdma buffers");
                }

                if (posix_memalign
                    ((void **) &c->RDMA_recv_buf_DMA, pagesize,
                     viadev_vbuf_total_size * viadev_num_rdma_buffer)) {
                    error_abort_all(GEN_EXIT_ERR,
                                    "Unable to malloc rdma buffers");
                }

                memset(c->RDMA_send_buf_DMA, 0,
                       viadev_vbuf_total_size * viadev_num_rdma_buffer);
                memset(c->RDMA_recv_buf_DMA, 0,
                       viadev_vbuf_total_size * viadev_num_rdma_buffer);

                /* REGISTER RDMA SEND BUFFERS */
                c->RDMA_send_buf_hndl =
                    register_memory(c->RDMA_send_buf_DMA,
                                    viadev_vbuf_total_size *
                                    viadev_num_rdma_buffer);

                /* REGISTER RDMA RECIEVE BUFFERS */
                c->RDMA_recv_buf_hndl =
                    register_memory(c->RDMA_recv_buf_DMA,
                                    viadev_vbuf_total_size *
                                    viadev_num_rdma_buffer);

                /* Connect the DMA buffer to the vbufs */
                for (k = 0; k < viadev_num_rdma_buffer; k++) {
                    /* Send vbufs */
                    v = &(c->RDMA_send_buf[k]);
                    v->head_flag =
                        (VBUF_FLAG_TYPE *) ((char *) (c->RDMA_send_buf_DMA)
                                            +
                                            (k * viadev_vbuf_total_size));
                    v->buffer =
                        (char *) ((char *) (v->head_flag) +
                                  sizeof(VBUF_FLAG_TYPE));

                    /* Receive vbufs */
                    v = &(c->RDMA_recv_buf[k]);
                    v->head_flag =
                        (VBUF_FLAG_TYPE *) ((char *) (c->RDMA_recv_buf_DMA)
                                            +
                                            (k * viadev_vbuf_total_size));
                    v->buffer =
                        (char *) ((char *) (v->head_flag) +
                                  sizeof(VBUF_FLAG_TYPE));
                }



                SAVE_MALLOC_HOOKS;
                SET_MVAPICH_MALLOC_HOOKS;

                /* set pointers */
                c->phead_RDMA_send = 0;
                c->ptail_RDMA_send = viadev_num_rdma_buffer - 1;

                c->p_RDMA_recv = 0;
                c->p_RDMA_recv_tail = viadev_num_rdma_buffer - 1;
#endif

            } else {
                c->phead_RDMA_send = 0;
                c->ptail_RDMA_send = 0;
                c->p_RDMA_recv = 0;
                c->p_RDMA_recv_tail = 0;
            }
            /* remote address and handle not available yet */
            c->remote_address_received = 0;

        }
#endif                          /* RDMA_FAST_PATH */

    }

#if defined(MCST_SUPPORT) && defined(MINISM)
    if (is_mcast_enabled) {
        /* Preposting the descriptors for the UD transport */
        for (i = 0; i < VIADEV_UD_PREPOST_DEPTH; i++) {
            PREPOST_UD_VBUF_RECV();
            ud_prepost_count++;
        }
    }
#endif                          /* MCST_SUPPORT && MINISM */

    /* call barrier to make sure all preposting has been done */
    D_PRINT("about to call pmgr_barrier...\n");

    /* call barrier to make sure all preposting has been done 
     * to gurantee all have preposted requests before real comm.
     * With MPD support, we use the BOOTSTRAP_CHANNEL,
     * which makes it simple to ensure the correctness of 
     * this barrier w.r.t to other traffic */
    i = pmgr_barrier();

    D_PRINT("process manager barrier returns %d\n", i);

    if (viadev_num_rdma_buffer == 0)
        return;

#if (defined(RDMA_FAST_PATH) || defined(MCST_SUPPORT))
    /* inform others about the receive 
     * RDMA buffer address and handle */

    for (i = 0; i < viadev.np; i++) {
        vbuf *v;
        viadev_packet_rdma_address *h;
        viadev_connection_t *c = &viadev.connections[i];

        if (i == viadev.me) {
            continue;
        }

        v = get_vbuf();
        h = (viadev_packet_rdma_address *) VBUF_BUFFER_START(v);

        /* set up the packet */
        PACKET_SET_HEADER(h, c, VIADEV_RDMA_ADDRESS);
        h->envelope.src_lrank = i;

#if defined(RDMA_FAST_PATH)
        /* set address and handle */
        h->RDMA_address = c->RDMA_recv_buf_DMA;
        h->RDMA_hndl = c->RDMA_recv_buf_hndl->rkey;
#endif                          /* RDMA_FAST_PATH */

#ifdef MCST_SUPPORT
        if (is_mcast_enabled) {
            /* set address and handle for UD Acks, this is for now remove later */
            h->RDMA_ACK_address = (viadev.bcast_info.ack_buffer);
            h->RDMA_ACK_hndl = viadev.bcast_info.ack_mem_hndl->rkey;
        }
#endif                          /* MCST_SUPPORT */

        /* finally send it off */
        vbuf_init_send(v, sizeof(viadev_packet_rdma_address));

        viadev_post_send(c, v);
    }

    /* get all the address */
    for (i = 0; i < viadev.np; i++) {
        viadev_connection_t *c = &viadev.connections[i];

        if (i == viadev.me) {
            continue;
        }

        while (!c->remote_address_received) {
            MPID_DeviceCheck(1);
        }
    }

#endif                          /* RDMA_FAST_PATH || MCST_SUPPORT */

}
#endif /* !ON_DEMAND */

#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))
void process_remote_rdma_address(viadev_connection_t * c,
                                 uint32_t remote_hndl,
                                 struct vbuf *remote_address);

void process_remote_rdma_address(viadev_connection_t * c,
                                 uint32_t remote_hndl,
                                 struct vbuf *remote_address)
{
    int i;

    c->remote_address_received = 1;

#ifndef ADAPTIVE_RDMA_FAST_PATH
    c->remote_RDMA_buf_hndl = remote_hndl;
    c->remote_RDMA_buf = remote_address;

    for (i = 0; i < viadev_num_rdma_buffer; i++) {

        /* set global_rank */
        c->RDMA_send_buf[i].grank = c->global_rank;
        c->RDMA_recv_buf[i].grank = c->global_rank;

        c->RDMA_recv_buf[i].desc.next = NULL;

        /* fill out the descriptors */
        c->RDMA_send_buf[i].padding = FREE_FLAG;

        /* Mark initialization state of RDMA buffers to be busy */
        c->RDMA_recv_buf[i].padding = BUSY_FLAG;

        /* this field to be filled out later */
        c->RDMA_send_buf[i].desc.sg_entry.length = 0;

        c->RDMA_send_buf[i].desc.sg_entry.addr =
            (uintptr_t) (c->RDMA_send_buf[i].head_flag);

        c->RDMA_send_buf[i].desc.sg_entry.lkey =
            c->RDMA_send_buf_hndl->lkey;

        c->RDMA_send_buf[i].desc.u.sr.sg_list =
            &(c->RDMA_send_buf[i].desc.sg_entry);

        c->RDMA_send_buf[i].desc.u.sr.send_flags = IBV_SEND_SIGNALED;
        c->RDMA_send_buf[i].desc.u.sr.opcode = IBV_WR_RDMA_WRITE;
        c->RDMA_send_buf[i].desc.u.sr.wr_id =
            (aint_t) & (c->RDMA_send_buf[i]);
        c->RDMA_send_buf[i].desc.u.sr.wr.rdma.rkey =
            c->remote_RDMA_buf_hndl;
        c->RDMA_send_buf[i].desc.u.sr.wr.rdma.remote_addr =
            (uintptr_t) ((char *) (c->remote_RDMA_buf) +
                         (i * viadev_vbuf_total_size));
        c->RDMA_send_buf[i].desc.u.sr.num_sge = 1;

    }
#else

    if (remote_address != 0) {

        /* Allocating the send vbufs for the eager RDMA flow */
        vbuf_fast_rdma_alloc(c, 0);
        c->remote_RDMA_buf_hndl = remote_hndl;
        c->remote_RDMA_buf = remote_address;

        for (i = 0; i < viadev_num_rdma_buffer; i++) {

            /* fill out the descriptors */

            /* this field to be filled out later */
            c->RDMA_send_buf[i].desc.sg_entry.length = 0;

            c->RDMA_send_buf[i].desc.sg_entry.addr =
                (uintptr_t) (c->RDMA_send_buf[i].head_flag);

            c->RDMA_send_buf[i].desc.sg_entry.lkey =
                c->RDMA_send_buf_hndl->lkey;

            c->RDMA_send_buf[i].desc.u.sr.sg_list =
                &(c->RDMA_send_buf[i].desc.sg_entry);

            c->RDMA_send_buf[i].padding = FREE_FLAG;

            c->RDMA_send_buf[i].desc.u.sr.send_flags = IBV_SEND_SIGNALED;
            c->RDMA_send_buf[i].desc.u.sr.opcode = IBV_WR_RDMA_WRITE;
            c->RDMA_send_buf[i].desc.u.sr.wr_id =
                (aint_t) & (c->RDMA_send_buf[i]);
            c->RDMA_send_buf[i].desc.u.sr.wr.rdma.rkey =
                c->remote_RDMA_buf_hndl;
            c->RDMA_send_buf[i].desc.u.sr.wr.rdma.remote_addr =
                (uintptr_t) ((char *) (c->remote_RDMA_buf) +
                             (i * viadev_vbuf_total_size));
            c->RDMA_send_buf[i].desc.u.sr.num_sge = 1;

        }
    }
#endif                          /* !ADAPTIVE_RDMA_FAST_PATH */

}

#endif

static void ib_finalize()
{
    int i;

#ifdef MCST_SUPPORT

    if (is_mcast_enabled) {
        ud_cleanup();
    }
#ifdef _SMP_
    if (smpi.my_local_id == 0) {
        leave_mcgrp(viadev_default_port);
    }
#else
    leave_mcgrp(viadev.hca_port_active);
#endif

#endif




#if (defined(RDMA_FAST_PATH) || defined(ADAPTIVE_RDMA_FAST_PATH))

    if (viadev_num_rdma_buffer) {
        /* unpin rdma buffers */
        for (i = 0; i < viadev.np; i++) {

            viadev_connection_t *c = &viadev.connections[i];

            if (i == viadev.me) {
                continue;
            }
#ifdef ON_DEMAND
            if (MPICM_IB_RC_PT2PT != MPICM_Get_conn_type(i)) {
                continue;
            }
#endif

            if (c->RDMA_send_buf_DMA) {
                if (ibv_dereg_mr(c->RDMA_send_buf_hndl)) {
                    error_abort_all(IBV_RETURN_ERR,
                                    "could not unpin send rdma buffer");
                }
            }

            if (c->RDMA_recv_buf_DMA) {
                if (ibv_dereg_mr(c->RDMA_recv_buf_hndl)) {
                    error_abort_all(IBV_RETURN_ERR,
                                    "could not unpin recv rdma buffer");
                }
            }

            /* free the buffers */
            if (c->RDMA_send_buf)
                free(c->RDMA_send_buf);
            if (c->RDMA_recv_buf)
                free(c->RDMA_recv_buf);
            if (c->RDMA_send_buf_DMA)
                free(c->RDMA_send_buf_DMA);
            if (c->RDMA_recv_buf_DMA)
                free(c->RDMA_recv_buf_DMA);

        }
    }
#endif

    if (NULL == viadev.qp_hndl) {
        error_abort_all(GEN_EXIT_ERR, "Null queue pair handle");
    }
#ifdef ON_DEMAND
    for (i = 0; i < viadev.np; i++) {
        if (MPICM_IB_RC_PT2PT == MPICM_Get_conn_type(i)) {
            if (ibv_destroy_qp(viadev.qp_hndl[i])) {
                error_abort_all(IBV_RETURN_ERR, "could not destroy QP");
            }
        }
    }
    MPICM_Finalize_UD();
    if (viadev.ud_qpn_table)
        free(viadev.ud_qpn_table);
    if (viadev.pending_req_head)
        free(viadev.pending_req_head);
    if (viadev.pending_req_tail)
        free(viadev.pending_req_tail);
    if (viadev.conn_table)
        free(viadev.conn_table);
#else
    for (i = 0; i < viadev.np; i++) {
        if (i == viadev.me) {
            continue;
        }
        if (ibv_destroy_qp(viadev.qp_hndl[i])) {
            error_abort_all(IBV_RETURN_ERR, "could not destroy QP");
        }
    }
#endif

    free(viadev.qp_hndl);

#ifdef USE_MPD_RING

    if (ibv_destroy_qp(viadev.boot_qp_hndl[0])) {
        error_abort_all(IBV_RETURN_ERR, "could not destroy lhs QP");
    }

    if (ibv_destroy_qp(viadev.boot_qp_hndl[1])) {
        error_abort_all(IBV_RETURN_ERR, "could not destroy rhs QP");
    }

    if (ibv_destroy_cq(viadev.boot_cq_hndl)) {
        error_abort_all(IBV_RETURN_ERR, "could not destroy CQ");
    }
#endif

#ifdef SRQ
    /* Cancel thread if active */
    pthread_cancel(viadev.async_thread);

    if (ibv_destroy_srq(viadev.srq_hndl)) {
        error_abort_all(IBV_RETURN_ERR, "Couldn't destroy SRQ\n");
    }
#endif

    {
        int ret = ibv_destroy_cq(viadev.cq_hndl);
        if (ret) {
            perror("ibv_destroy_cq");
            error_abort_all(IBV_RETURN_ERR,
                            "could not destroy CQ, return value=%d\n",
                            ret);
        }

    }

    deallocate_vbufs();

    /* unregister all user buffer registration */
    while (dreg_evict());

    if (ibv_dealloc_pd(viadev.ptag)) {
        error_abort_all(IBV_RETURN_ERR, "could not dealloc PD");
    }

    /* to release all resources */
    if (ibv_close_device(viadev.context)) {
        error_abort_all(IBV_RETURN_ERR, "could not close device");
    }
}

int rank_to_hca_lid(int i)
{
    return viadev.lid_table[i];
}

int rank_to_qp_num(int i)
{
    return viadev.qp_table[i];
}


#if defined(_SMP_) || defined(USE_MPD_BASIC) || defined(USE_MPD_RING)
/* gethostname does return a NULL-terminated string if
 * the hostname length is less than "HOSTNAME_LEN". Otherwise,
 * it is unspecified.
 */
static int get_host_id(char *myhostname, int hostname_len)
{
    int host_id = 0;
    struct hostent *hostent;

    hostent = gethostbyname(myhostname);
    if (hostent == NULL) {
        error_abort_all(GEN_EXIT_ERR,
                        "Null value returned by gethostbyname");
    }

    host_id = ((struct in_addr *) hostent->h_addr_list[0])->s_addr;

    return host_id;
}

#endif                          /* SMP */

#ifndef ON_DEMAND
static void ib_rank_lid_table(int *alladdrs, int local_addr_len)
{
    int i;

#ifdef _SMP_
    int *hostnames_j = NULL;
    int *smpi_ptr = NULL;
    int j;

#ifdef _AFFINITY_
    long N_CPUs_online;
    unsigned long affinity_mask = 1;
    unsigned long affinity_mask_len = sizeof(affinity_mask);

    /*Get number of CPU on machine */
    if ((N_CPUs_online = sysconf(_SC_NPROCESSORS_ONLN)) < 1) {
        perror("sysconf");
    }
#endif /*_AFFINITY_*/

    SET_ORIGINAL_MALLOC_HOOKS;

    hostnames_j = (int *) malloc(viadev.np * sizeof(int));
    smpi.local_nodes = (unsigned int *) malloc(viadev.np * sizeof(int));
    smpi_ptr = (unsigned int *) malloc(viadev.np * sizeof(int));

    if (hostnames_j == NULL || smpi.local_nodes == NULL
        || smpi_ptr == NULL) {
        error_abort_all(GEN_EXIT_ERR,
                        "malloc: in ib_rank_lid_table for SMP");
    }

    SAVE_MALLOC_HOOKS;
    SET_MVAPICH_MALLOC_HOOKS;
#endif

#if defined(USE_MPD_BASIC) || defined (USE_MPD_RING)
    for (i = 0; i < viadev.np; i++) {

        char *alladdr_inv;
        char *temp_ptr;
        char temp_lid[12];
        char temp_qp[12];
        char temp_hostid[12];

        alladdr_inv =
            (char *) alladdrs + i * ((viadev.np + 1) * QPLEN_XDR);

        {
            /* Get the qp # */
            temp_ptr = alladdr_inv + viadev.me * QPLEN_XDR;
            strncpy(temp_qp, temp_ptr, QPLEN_XDR);
            temp_qp[QPLEN_XDR] = '\0';

            /* Get hca lid # */
            temp_ptr = alladdr_inv + i * QPLEN_XDR;
            strncpy(temp_lid, temp_ptr, QPLEN_XDR);
            temp_lid[QPLEN_XDR] = '\0';

            /* Get host id # */
            temp_ptr = alladdr_inv + viadev.np * QPLEN_XDR;
            strncpy(temp_hostid, temp_ptr, QPLEN_XDR);
            temp_hostid[QPLEN_XDR] = '\0';
        }

        if (i == viadev.me) {
            /* No QP is allocated for a process itself 
             * If you change this, please change mpirun_rsh.c:684 too */
            viadev.qp_table[i] = -1;
        } else {
            viadev.qp_table[i] = atoi(temp_qp);
        }
        viadev.lid_table[i] = atoi(temp_lid);
        D_PRINT("rank_lid[%d]: hca_lid=%d, qp_num=%d\n",
                i, viadev.lid_table[i], viadev.qp_table[i]);
#ifdef _SMP_
        sscanf(temp_hostid, "%08X", &(hostnames_j[i]));
#endif
    }
#else                           /* (defined(USE_MPD_BASIC) || defined (USE_MPD_RING)) */

    for (i = 0; i < viadev.np; i++) {
        viadev.lid_table[i] = alladdrs[i];
        viadev.qp_table[i] = alladdrs[i + viadev.np];
        D_PRINT("rank_lid[%d]: hca_lid=%d, qp_num=%d\n",
                i, viadev.lid_table[i], viadev.qp_table[i]);

#ifdef _SMP_
        hostnames_j[i] = alladdrs[2 * viadev.np + i];
#endif
    }
#endif                          /* End of ! (defined(USE_MPD_BASIC) || defined (USE_MPD_RING)) */

#ifdef _SMP_

    smpi.only_one_device = 1;
    smpi.num_local_nodes = 0;
    for (j = 0; j < viadev.np; j++) {
        if (hostnames_j[viadev.me] == hostnames_j[j]) {
            if (j == viadev.me) {
                smpi.my_local_id = smpi.num_local_nodes;
#ifdef _AFFINITY_
                if (viadev_enable_affinity > 0) {
                    affinity_mask =
                        1 << (smpi.my_local_id % N_CPUs_online);
                    if (sched_setaffinity
                        (0, affinity_mask_len, &affinity_mask) < 0) {
                        perror("sched_setaffinity");
                    }
                }
#endif                          /* _AFFINITY_ */
            }
            smpi.local_nodes[j] = smpi.num_local_nodes;
            smpi_ptr[smpi.num_local_nodes] = j;
            smpi.num_local_nodes++;
        } else {
            smpi.only_one_device = 0;
            smpi.local_nodes[j] = -1;
        }
    }
    smpi.l2g_rank =
        (unsigned int *) malloc(smpi.num_local_nodes * sizeof(int));
    if (smpi.l2g_rank == NULL) {
        error_abort_all(GEN_EXIT_ERR,
                        "malloc: in ib_rank_lid_table for SMP");
    }
    for (j = 0; j < smpi.num_local_nodes; j++) {
        smpi.l2g_rank[j] = smpi_ptr[j];
    }
    free(smpi_ptr);
    free(hostnames_j);
#endif
}
#endif /* !ON_DEMAND */

#ifdef USE_MPD_RING

/* A ring-based barrier: process 0 initiates two tokens 
 * going clockwise and counter-clockwise, respectively.
 * The return of these two tokens completes the barrier.
 */
static int bootstrap_barrier(void)
{
    /* work entries related variables */
    struct ibv_send_wr sr;
    struct ibv_sge sg_entry_s;
    struct ibv_send_wr *bad_wr_s;

    /* completion related variables */
    struct ibv_cq *cq_hndl;
    struct ibv_wc rc;

    /* Memory related variables */
    struct ibv_mr *mem_handle;
    int unit_length;
    char *send_addr;
    char *temp_buff;
    int barrier_recvd = 0;
    int ne;

    /* Find out the barrier address */
    unit_length = QPLEN_XDR * (viadev.np + 1);
    temp_buff = (char *) barrier_buff + viadev.me * unit_length;
    mem_handle = barrier_mem_hndl;

    sr.opcode = IBV_WR_SEND;
    sr.send_flags = IBV_SEND_SIGNALED;
    sr.num_sge = 1;
    sr.sg_list = &sg_entry_s;
    sr.next = NULL;
    sg_entry_s.lkey = mem_handle->lkey;
    cq_hndl = viadev.boot_cq_hndl;

    if (viadev.me == 0) {
        /* send_lhs(); send_rhs(); recv_lhs(); recv_rhs(); */
        sr.wr_id = viadev.np * 3 + 0;
        sg_entry_s.length = QPLEN_XDR;
        send_addr = temp_buff + unit_length;
        snprintf(send_addr, QPLEN_XDR, "000");
        sg_entry_s.addr = (uintptr_t) send_addr;

        if (ibv_post_send(viadev.boot_qp_hndl[1], &sr, &bad_wr_s)) {
            error_abort_all(IBV_RETURN_ERR, "Error posting send!\n");
        }

        sr.wr_id = viadev.np * 3 + 1;
        sg_entry_s.length = 2 * QPLEN_XDR;
        send_addr = temp_buff + unit_length + QPLEN_XDR;
        snprintf(send_addr, 2 * QPLEN_XDR, "111");
        sg_entry_s.addr = (uintptr_t) send_addr;

        if (ibv_post_send(viadev.boot_qp_hndl[0], &sr, &bad_wr_s)) {
            error_abort_all(IBV_RETURN_ERR, "Error posting send!\n");
        }

        while (barrier_recvd < 2) {

            /* Polling CQ */
            ne = ibv_poll_cq(cq_hndl, 1, &rc);
            if (ne < 0) {
                error_abort_all(IBV_RETURN_ERR, "Poll CQ failed!\n");
            } else if (ne > 1) {
                error_abort_all(IBV_RETURN_ERR, "Got more than one\n");
            } else if (ne == 1) {
                D_PRINT("** Receive forward %s reverse %s id %d \n",
                        (char *) temp_buff,
                        (char *) (temp_buff + QPLEN_XDR), (int) rc.wr_id);
                if (rc.status != IBV_WC_SUCCESS) {
                    error_abort_all(IBV_RETURN_ERR,
                                    " [%s:%d] Error code in %d polled desc!\n",
                                    viadev.my_name, viadev.me, rc.status);
                }
                /* Make sure it is a recv completion */
                else if (rc.opcode) {
                    D_PRINT("Receive forward %s reverse %s id %d \n",
                            (char *) temp_buff,
                            (char *) (temp_buff + QPLEN_XDR),
                            (int) rc.wr_id);
                    barrier_recvd++;
                } else {
                    D_PRINT("Received code (s): %d, id: %d\n", rc.opcode,
                            (int) rc.wr_id);
                }
            }
        }                       /* end of while loop */
    } else {
        /* recv_lhs(); send_rhs(); recv_rhs(); send_lhs(); */

        while (barrier_recvd < 2) {
            /* Polling CQ */
            ne = ibv_poll_cq(cq_hndl, 1, &rc);
            if (ne < 0) {
                error_abort_all(IBV_RETURN_ERR, "Poll CQ failed!\n");
            } else if (ne > 1) {
                error_abort_all(IBV_RETURN_ERR, "Got more than one\n");
            } else if (ne == 1) {
                D_PRINT("** Receive forward %s reverse %s id %d \n",
                        (char *) temp_buff,
                        (char *) (temp_buff + QPLEN_XDR), (int) rc.wr_id);
                if (rc.status != IBV_WC_SUCCESS) {
                    error_abort_all(IBV_RETURN_ERR,
                                    " [%s:%d] Error code %d in polled desc!\n",
                                    viadev.my_name, viadev.me, rc.status);
                }
                /* Make sure it is a recv completion */
                else if (rc.opcode) {
                    D_PRINT("Received forward %s reverse %s id %d\n",
                            (char *) temp_buff,
                            (char *) (temp_buff + QPLEN_XDR),
                            (int) rc.wr_id);
                    barrier_recvd++;

                    /* find out which one is received, and 
                     * forward accordingly */
                    sr.wr_id = rc.wr_id + 1 - viadev.np;
                    sr.next = NULL;
                    sg_entry_s.length = (sr.wr_id + 1) * QPLEN_XDR;
                    send_addr = temp_buff + sr.wr_id * QPLEN_XDR;
                    sg_entry_s.addr = (uintptr_t) send_addr;

                    if (ibv_post_send(viadev.boot_qp_hndl[1 - sr.wr_id],
                                      &sr, &bad_wr_s)) {
                        error_abort_all(IBV_RETURN_ERR,
                                        "Error posting second send!\n");
                    }
                } /* end else if (rc.opcode == IBV_WR_SEND) */
                else {
                    D_PRINT("Received code (r): %d id: %d\n", rc.opcode,
                            (int) rc.wr_id);
                }
            }                   /* end else if (ne == 1)                  */
        }                       /* end while                              */
    }                           /* end else                               */
    D_PRINT("Done with barrier\n");
    return (1);
}

/* Set up a ring of qp's, here a separate bootstrap channel is used.*/
static void ib_vapi_enable_ring(int lhs, int rhs, char *ring_addr)
{
    struct ibv_qp_attr qp_attr;

    int i;

    /* Get lhs and rhs hca_lid and qp # */
    char temp_str[QPLEN_XDR + 1];
    char *temp_ptr;

    temp_str[QPLEN_XDR] = '\0';

    for (i = 0; i < 2; i++) {
        /* For hca_lid */
        temp_ptr = ring_addr + i * (QPLEN_XDR * 3);
        strncpy(temp_str, temp_ptr, QPLEN_XDR);
        D_PRINT("Got hca_lid %s \n", temp_str);
        viadev.boot_tb[i][0] = atoi(temp_str);

        temp_ptr += QPLEN_XDR * (2 - i);

        /* qp # to me usually on rhs unless I am at the beginning */
        strncpy(temp_str, temp_ptr, QPLEN_XDR);
        D_PRINT("Got queue_pair %s \n", temp_str);
        viadev.boot_tb[i][1] = atoi(temp_str);
    }

    /* Modifying QP to INIT */

    memset(&qp_attr, 0, sizeof(qp_attr));
    qp_attr.qp_state = IBV_QPS_INIT;
    qp_attr.pkey_index = 0;
    qp_attr.port_num = viadev_default_port;
    qp_attr.qp_access_flags = IBV_ACCESS_REMOTE_WRITE |
        IBV_ACCESS_LOCAL_WRITE;

    if (ibv_modify_qp(viadev.boot_qp_hndl[0], &qp_attr,
                      IBV_QP_STATE |
                      IBV_QP_PKEY_INDEX |
                      IBV_QP_PORT | IBV_QP_ACCESS_FLAGS)) {
        error_abort_all(IBV_RETURN_ERR,
                        "Could not modify lhs qp to INIT!\n");
    }

    if (ibv_modify_qp(viadev.boot_qp_hndl[1], &qp_attr,
                      IBV_QP_STATE |
                      IBV_QP_PKEY_INDEX |
                      IBV_QP_PORT | IBV_QP_ACCESS_FLAGS)) {
        error_abort_all(IBV_RETURN_ERR,
                        "Could not modify rhs qp to INIT!\n");
    }

    /* INIT -> RTR */

    memset(&qp_attr, 0, sizeof(qp_attr));
    qp_attr.qp_state = IBV_QPS_RTR;
    qp_attr.path_mtu = viadev_default_mtu;
    qp_attr.rq_psn = viadev_default_psn;
    qp_attr.max_dest_rd_atomic = viadev_default_qp_ous_rd_atom;
    qp_attr.min_rnr_timer = viadev_default_min_rnr_timer;

    qp_attr.ah_attr.is_global = 0;
    qp_attr.ah_attr.sl = viadev_default_service_level;
    qp_attr.ah_attr.src_path_bits = viadev_default_src_path_bits;
    qp_attr.ah_attr.port_num = viadev_default_port;
    qp_attr.ah_attr.static_rate = viadev_default_static_rate;

    for (i = 0; i < 2; i++) {
        qp_attr.dest_qp_num = viadev.boot_tb[i][1];
        qp_attr.ah_attr.dlid = viadev.boot_tb[i][0];

        D_PRINT("Remote LID[%d] QP=%d\n", qp_attr.ah_attr.dlid,
                qp_attr.dest_qp_num);
        if (ibv_modify_qp(viadev.boot_qp_hndl[i], &qp_attr,
                          IBV_QP_STATE |
                          IBV_QP_PATH_MTU |
                          IBV_QP_RQ_PSN |
                          IBV_QP_MAX_DEST_RD_ATOMIC |
                          IBV_QP_MIN_RNR_TIMER |
                          IBV_QP_DEST_QPN | IBV_QP_AV)) {
            error_abort_all(GEN_EXIT_ERR, "Failed to modify QP to RTR\n");
        }

        D_PRINT("Modified to RTR..Qp[%d]\n", i);
    }

    /* RTR -> RTS */

    memset(&qp_attr, 0, sizeof(qp_attr));
    qp_attr.qp_state = IBV_QPS_RTS;
    qp_attr.sq_psn = viadev_default_psn;
    qp_attr.timeout = viadev_default_time_out;
    qp_attr.retry_cnt = viadev_default_retry_count;
    qp_attr.rnr_retry = viadev_default_rnr_retry;
    qp_attr.max_rd_atomic = viadev_default_qp_ous_rd_atom;

    {
        if (ibv_modify_qp(viadev.boot_qp_hndl[0], &qp_attr,
                          IBV_QP_STATE |
                          IBV_QP_SQ_PSN |
                          IBV_QP_TIMEOUT |
                          IBV_QP_RETRY_CNT |
                          IBV_QP_RNR_RETRY | IBV_QP_MAX_QP_RD_ATOMIC)) {
            error_abort_all(GEN_EXIT_ERR,
                            "Failed to modify LHS QP to RTS\n");
        }

        if (ibv_modify_qp(viadev.boot_qp_hndl[1], &qp_attr,
                          IBV_QP_STATE |
                          IBV_QP_SQ_PSN |
                          IBV_QP_TIMEOUT |
                          IBV_QP_RETRY_CNT |
                          IBV_QP_RNR_RETRY | IBV_QP_MAX_QP_RD_ATOMIC)) {
            error_abort_all(GEN_EXIT_ERR,
                            "Failed to modify RHS QP to RTS\n");
        }

        D_PRINT("Modified to RTS..\n");
    }
}

/* Using a ring of queue pairs to exchange all the queue_pairs,
 * If mpd is used,  only info about lhs and rsh are provided. */
static void ib_vapi_bootstrap_ring(int lhs, int rhs,
                                   char *ring_addr, char *local_addr,
                                   char *alladdrs,
                                   struct ibv_mr *mem_handle)
{
    int j;
    int ret;
    int ne;
    int recv_index;
    int send_comp, recv_comp;

    /* Now enable the queue pairs and post descriptors */
    ib_vapi_enable_ring(lhs, rhs, ring_addr);

    /* Register alladdrs and post receive descriptors */
    {
        /* work entries related variables */
        struct ibv_recv_wr rr;
        struct ibv_sge sg_entry_r;
        struct ibv_recv_wr *bad_wr_r;
        struct ibv_send_wr sr;
        struct ibv_sge sg_entry_s;
        struct ibv_send_wr *bad_wr_s;

        /* completion related variables */
        struct ibv_cq *cq_hndl;
        struct ibv_wc rc;

        int unit_length;
        char *dest_loc;
        char *recv_addr;
        char *send_addr;

        /* Same as local_addr_len */
        unit_length = (viadev.np + 1) * QPLEN_XDR;

        /* Copy local_addr to the correct slot in alladdrs
         * and  post receive descriptors for all_addr */
        dest_loc = alladdrs + viadev.me * unit_length;
        strncpy(dest_loc, local_addr, unit_length);

        recv_index = 0;

        /* Post receive for all_addr */
        for (j = 0; j < viadev.np + 1; j++) {
            /* The last two entries are used for a barrier,
             * they overlap the local address */
            if ((j + 1) < viadev.np) {
                recv_addr = alladdrs + unit_length *
                    ((viadev.me + viadev.np - j - 1) % viadev.np);
            } else if (j < viadev.np) {
                recv_addr = alladdrs + unit_length * viadev.me;
            } else {
                recv_addr = alladdrs + unit_length * viadev.me + QPLEN_XDR;
            }

            /* Fillup a recv descriptor */

            rr.wr_id = j;
            rr.num_sge = 1;
            rr.sg_list = &(sg_entry_r);
            rr.next = NULL;
            sg_entry_r.lkey = mem_handle->lkey;
            sg_entry_r.addr = (uintptr_t) recv_addr;

            if ((j + 1) >= viadev.np)
                sg_entry_r.length = (j + 2 - viadev.np) * QPLEN_XDR;
            else
                sg_entry_r.length = unit_length;

            /* Post the recv descriptor */
            if (j < viadev.np) {
                ret =
                    ibv_post_recv(viadev.boot_qp_hndl[0], &rr, &bad_wr_r);
            } else {
                ret =
                    ibv_post_recv(viadev.boot_qp_hndl[1], &rr, &bad_wr_r);
            }
            if (ret) {
                error_abort_all(IBV_RETURN_ERR, "Error posting recv!\n");
            }
        }

        /* synchronize all the processes */
        pmgr_barrier();

        recv_index = 0;
        recv_comp = 0;
        /* transfer all the addresses */
        for (j = 0; j < viadev.np - 1; j++) {

            D_PRINT("j = %d, np: %d\n", j, viadev.np);
            send_addr = alladdrs +
                unit_length * ((viadev.me + viadev.np - j) % viadev.np);

            /* send to rhs */
            sr.opcode = IBV_WR_SEND;
            sr.send_flags = IBV_SEND_SIGNALED;
            sr.wr_id = j + 2 * viadev.np;
            sr.num_sge = 1;
            sr.sg_list = &sg_entry_s;
            sr.next = NULL;
            sg_entry_s.addr = (uintptr_t) send_addr;
            sg_entry_s.length = unit_length;
            sg_entry_s.lkey = mem_handle->lkey;

            if (ibv_post_send(viadev.boot_qp_hndl[1], &sr, &bad_wr_s)) {
                error_abort_all(IBV_RETURN_ERR, "Error posting send!\n");
            }

            /* recv from lhs */
            cq_hndl = viadev.boot_cq_hndl;

            rc.wr_id = -1;
            send_comp = -1;
            do {
                ne = ibv_poll_cq(cq_hndl, 1, &rc);
                if (ne < 0) {
                    error_abort_all(IBV_RETURN_ERR, "Poll CQ failed!\n");
                } else if (ne > 1) {
                    error_abort_all(IBV_RETURN_ERR, "Got more than one\n");
                } else if (ne == 1) {
                    if (rc.status != IBV_WC_SUCCESS) {
                        error_abort_all(IBV_RETURN_ERR,
                                        " [%s:%d] Error code %d in polled desc!\n",
                                        viadev.my_name, viadev.me,
                                        rc.status);
                    }

                    D_PRINT
                        ("Sucessful Poll rc.opcode %d, rc.wr_id is %d\n",
                         rc.opcode, rc.wr_id);

                    if (rc.wr_id == (j + 2 * viadev.np))
                        send_comp = rc.wr_id;
                    else
                        recv_comp = rc.wr_id + 1;

                    D_PRINT("send_comp: %d, recv_comp: %d\n", send_comp,
                            recv_comp);

                    if (rc.opcode == IBV_WR_SEND) {
                        if (rc.wr_id != recv_index) {
                            fprintf(stderr,
                                    "[%d](%s:%d)unexpected message %d vs %d\n",
                                    viadev.me, __FILE__, __LINE__,
                                    (int) rc.wr_id, recv_index);
                            exit(1);
                        } else {
                            D_PRINT("expected message %d\n", rc.wr_id);
                            recv_index = rc.wr_id + 1;
                            D_PRINT("recv_index %d\n", recv_index);
                        }
                    }
                }
            } while (!(send_comp == (j + 2 * viadev.np) && recv_comp > j));     /* do-while loop */
            D_PRINT
                ("Finished polling section: send_comp = %d, recv_comp = %d\n",
                 send_comp, recv_comp);
        }                       /* end of for loop */

    }                           /* end of alltoall exchange */
}
#endif

#ifdef MCST_SUPPORT

 /* Creation of a separate UD_QP and attaching the QP to the mcst_grp */
static void ib_ud_qp_init(void)
{
    struct ibv_qp_init_attr qp_init_attr;
    struct ibv_qp_attr qp_attr;
    int ret;

    viadev.ud_scq_hndl =
        ibv_create_cq(viadev.context, DEFAULT_MAX_CQ_SIZE, NULL, NULL, 0);

    if (!viadev.ud_scq_hndl) {
        error_abort_all(IBV_RETURN_ERR, "Cannot allocate CQ for UD-QP\n");
    }

    viadev.ud_rcq_hndl =
        ibv_create_cq(viadev.context, DEFAULT_MAX_CQ_SIZE, NULL, NULL, 0);

    if (!viadev.ud_rcq_hndl) {
        error_abort_all(IBV_RETURN_ERR, "Cannot allocate CQ for UD-QP");
    }

    memset(&qp_init_attr, 0, sizeof(qp_init_attr));

    qp_init_attr.cap.max_recv_wr = DEFAULT_MAX_WQE;
    qp_init_attr.cap.max_send_wr = DEFAULT_MAX_WQE;

    qp_init_attr.cap.max_recv_sge = viadev_default_max_sg_list;
    qp_init_attr.cap.max_send_sge = viadev_default_max_sg_list;
    qp_init_attr.cap.max_inline_data = 0;

    qp_init_attr.recv_cq = viadev.ud_rcq_hndl;
    qp_init_attr.send_cq = viadev.ud_scq_hndl;
    qp_init_attr.sq_sig_all = 0;
    qp_init_attr.qp_type = IBV_QPT_UD;

    viadev.ud_qp_hndl = ibv_create_qp(viadev.ptag, &qp_init_attr);

    if (!viadev.ud_qp_hndl) {
        error_abort_all(IBV_RETURN_ERR, "Cannot create UD-QP");
    }

    memset(&qp_attr, 0, sizeof(qp_attr));

    qp_attr.qp_state = IBV_QPS_INIT;
    qp_attr.pkey_index = 0;
    qp_attr.port_num = viadev_default_port;
    qp_attr.qkey = 0;
    if (ret = ibv_modify_qp(viadev.ud_qp_hndl, &qp_attr,
                            IBV_QP_STATE |
                            IBV_QP_PKEY_INDEX |
                            IBV_QP_PORT | IBV_QP_QKEY)) {
        printf("ret value:%d\n", ret);
        error_abort_all(IBV_RETURN_ERR, "Could not modify UD-QP to INIT");
    }

    D_PRINT("Modified UD to init..Qp\n");

    /**********************  INIT --> RTR  ************************/
    memset(&qp_attr, 0, sizeof(qp_attr));

    qp_attr.qp_state = IBV_QPS_RTR;

    if (ibv_modify_qp(viadev.ud_qp_hndl, &qp_attr, IBV_QP_STATE))
        error_abort_all(IBV_RETURN_ERR, "Could not modify UD-QP to RTR");

    D_PRINT("Modified UD to RTR..Qp\n");


    /********************* RTR --> RTS  * *******************/
    qp_attr.qp_state = IBV_QPS_RTS;
    qp_attr.sq_psn = viadev_default_psn;

    if (ibv_modify_qp(viadev.ud_qp_hndl, &qp_attr,
                      IBV_QP_STATE | IBV_QP_SQ_PSN))
        error_abort_all(IBV_RETURN_ERR, "Could not modify UD-QP to RTS");

    D_PRINT("Modified UD to RTS..Qp\n");

    /* Attach to the mulitcast grp - currently all nodes are in bcast grp */

    if (ibv_attach_mcast(viadev.ud_qp_hndl, &viadev.mcg_dgid, 0))
        error_abort_all(IBV_RETURN_ERR,
                        "Could not attach UD-QP to MCGrp\n");
}

static void ud_mcst_av_init(void)
{
    memset(&viadev.av, 0, sizeof(viadev.av));
    viadev.av.grh.dgid.raw[0] = 0;
    viadev.av.dlid = mlid;

    viadev.av.is_global = TRUE;
    viadev.av.grh.flow_label = DEFAULT_FLOW_LABEL;
    viadev.av.grh.hop_limit = DEFAULT_HOP_LIMIT;
    viadev.av.grh.sgid_index = 0;
    viadev.av.grh.traffic_class = DEFAULT_TRAFFIC_CLASS;
    memcpy(viadev.av.grh.dgid.raw, viadev.mcg_dgid.raw,
           sizeof(union ibv_gid));
    viadev.av.sl = viadev_default_service_level;
    viadev.av.port_num = viadev.hca_port_active;
    viadev.av.src_path_bits = viadev_default_src_path_bits;
    viadev.av.static_rate = viadev_default_static_rate;

    viadev.av_hndl = ibv_create_ah(viadev.ptag, &viadev.av);

    if (!viadev.av_hndl) {
        error_abort_all(IBV_RETURN_ERR,
                        "Could not create address vector\n");
    }
}


static void alloc_ack_region()
{

    int mem_total;
    int alignment = 4;          /* aligning to a 4 byte boundary */

    if (LEFT_OVERS == 0) {
        mem_total =
            (NUM_CHILD * NUM_COROOTS + NUM_COROOTS +
             NUM_WBUFFERS) * ACK_LEN + alignment - 1;
    } else {
        if (viadev.me == 0) {
            mem_total =
                ((NUM_CHILD + LEFT_OVERS) * NUM_COROOTS + NUM_COROOTS +
                 NUM_WBUFFERS) * ACK_LEN + alignment - 1;
        } else {
            if (viadev.me >= NUM_CHILD * NUM_COROOTS) {
                mem_total = (viadev.np + 1 + NUM_WBUFFERS) * ACK_LEN
                    + alignment - 1;
            } else {
                mem_total =
                    (NUM_CHILD * NUM_COROOTS + NUM_COROOTS +
                     NUM_WBUFFERS) * ACK_LEN + alignment - 1;
            }
        }

    }

    viadev.bcast_info.ack_region = malloc(mem_total);
    memset(viadev.bcast_info.ack_region, 0, mem_total);

    viadev.bcast_info.ack_buffer =
        (void
         *) (((aint_t) viadev.bcast_info.ack_region +
              (aint_t) (alignment - 1)) & ~((aint_t) alignment - 1));

    viadev.bcast_info.ack_mem_hndl =
        ibv_reg_mr(viadev.ptag, viadev.bcast_info.ack_buffer,
                   mem_total - alignment - 1,
                   IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE);

    co_print("local mem:%p\n", viadev.bcast_info.ack_buffer);
    if (!(viadev.bcast_info.ack_mem_hndl)) {
        error_abort_all(IBV_RETURN_ERR, "Couldn't register memory\n");
    }
    co_print("local r_key:%d\n", mr_out.r_key);
}

static void alloc_bcast_buffers(void)
{
    int i, j;
    int *ack_ptr = (int *) bcast_info->ack_buffer;
    int num_coroots, num_child;

    if (LEFT_OVERS == 0) {
        num_coroots = NUM_COROOTS;
        num_child = NUM_CHILD;
    } else {
        if (viadev.me > LAST_NODE) {
            num_coroots = 1;
            num_child = viadev.np;
        } else {
            if (viadev.me == 0) {
                num_coroots = NUM_COROOTS;
                num_child = NUM_CHILD + LEFT_OVERS;
            } else {
                num_coroots = NUM_COROOTS;
                num_child = NUM_CHILD;
            }
        }
    }

    bcast_info->ack_cnt = malloc((sizeof(int)) * (viadev.np));
    bcast_info->msg_cnt = malloc((sizeof(int)) * (viadev.np));
    bcast_info->time_out =
        (double **) malloc((sizeof(double *)) * num_coroots);
    bcast_info->win_head = malloc((sizeof(int)) * num_coroots);
    bcast_info->win_tail = malloc((sizeof(int)) * num_coroots);
    bcast_info->buf_head = malloc((sizeof(int)) * num_coroots);
    bcast_info->buf_tail = malloc((sizeof(int)) * num_coroots);
    bcast_info->min_ack = malloc((sizeof(int)) * num_coroots);
    bcast_info->is_full = malloc((sizeof(int)) * num_coroots);
    bcast_info->co_buf = malloc((sizeof(vbuf *)) * num_coroots);


    for (i = 0; i < num_coroots; i++) {
        bcast_info->win_head[i] = bcast_info->win_tail[i] = 0;
        bcast_info->buf_head[i] = bcast_info->buf_tail[i] = 0;
        bcast_info->sbuf_head = bcast_info->sbuf_tail = 0;
        bcast_info->ack_full = 0;
        bcast_info->is_full[i] = 0;
        bcast_info->co_buf[i] = malloc((sizeof(vbuf)) * SENDER_WINDOW);
        bcast_info->time_out[i] =
            (double *) malloc((sizeof(double)) * SENDER_WINDOW);
        bcast_info->min_ack[i] = -1;

    }

    for (i = 0; i < viadev.np; i++)
        bcast_info->Bcnt[i] = 0;

    for (i = 0; i < num_child * num_coroots + num_coroots + NUM_WBUFFERS;
         i++)
        ack_ptr[i] = -1;

    for (i = 0; i < num_coroots; i++) {
        for (j = 0; j < SENDER_WINDOW; j++)
            bcast_info->time_out[i][j] = -1;
    }

    bcast_info->bcast_called = 0;
}

void ud_cleanup(void)
{

    int i;
    int num_coroots;

    if (LEFT_OVERS == 0) {
        num_coroots = NUM_COROOTS;
    } else {
        if (viadev.me > LAST_NODE) {
            num_coroots = 1;
        } else {
            num_coroots = NUM_COROOTS;
        }
    }


    free(bcast_info->ack_cnt);
    free(bcast_info->msg_cnt);
    free(bcast_info->win_head);

    free(bcast_info->win_tail);
    free(bcast_info->buf_head);
    free(bcast_info->buf_tail);
    free(bcast_info->is_full);
    free(bcast_info->buf_credits);
    free(bcast_info->buf_updates);

    for (i = 0; i < num_coroots; i++) {
        free(bcast_info->co_buf[i]);
        free(bcast_info->time_out[i]);
    }
    free(bcast_info->time_out);
    free(bcast_info->co_buf);
    ibv_detach_mcast(viadev.ud_qp_hndl, &viadev.mcg_dgid, 0);
    ibv_destroy_qp(viadev.ud_qp_hndl);
    ibv_destroy_ah(viadev.av_hndl);
    ibv_destroy_cq(viadev.ud_scq_hndl);
    ibv_destroy_cq(viadev.ud_rcq_hndl);
    ibv_dereg_mr(bcast_info->ack_mem_hndl);
    free(bcast_info->ack_buffer);
}

#endif

#ifdef PENTIUM_MHZ
double get_us(void)
{
    unsigned long long now;
    __asm __volatile(".byte 0x0f, 0x31":"=A"(now));
    return (double) ((double) now / (PENTIUM_MHZ));
}
#else
double get_us(void)
{
    struct timeval t;
    gettimeofday(&t, NULL);
    return (double) t.tv_sec * (double) 1e6 + (double) t.tv_usec;
}
#endif
