/* Copyright (c) 2002-2006, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#ifndef _VAPI_HEADER_H
#define _VAPI_HEADER_H


#undef IN
#undef OUT

#include <infiniband/verbs.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <assert.h>

#include "ibverbs_const.h"
#include "via64.h"
#include "viaparam.h"

double get_us(void);

#define INVAL_HNDL (0xffffffff)


#define IN
#define OUT

#undef MALLOC
#undef FREE
/* src/env/initutil.c NEW not defined */
#define MALLOC(a)    malloc((unsigned)(a))
#define CALLOC(a,b)  calloc((unsigned)(a),(unsigned)(b))
#define FREE(a)      free((char *)(a))
#define NEW(a)    (a *)MALLOC(sizeof(a))
#define STRDUP(a) 	strdup(a)

#ifdef _PPC64_
#define stbar()  asm volatile ("sync": : :"memory")     /* ": : :" for C++ */
#define readbar() asm volatile ("lwsync": : :"memory")
#define writebar() asm volatile ("eieio": : :"memory")
#else
#define stbar()  
#define readbar() 
#define writebar() 
#endif

/* memory handle */
typedef struct {
	int a;
} VIP_MEM_HANDLE;

extern int backlog_counter;
/* for multiple HCAs/ports support */
#define MAX_NUM_HCAS    4 
#define MAX_NUM_PORTS   4
#define MAX_SUBCHANNELS (MAX_NUM_HCAS * MAX_NUM_PORTS)
#define STRIPING_THRESHOLD 6*1024

#if 0
extern uint16_t num_ports;
extern uint16_t num_hcas;
extern uint16_t num_subchannels;
#else
extern int num_ports;
extern int num_hcas;
extern int num_subchannels;
extern int num_phy_ports;
extern int num_qp_per_port;

extern int lm_scheduling;
#define EVEN_STRIPING 0
#define WEIGHTED_STRIPING 1
#define STRIPE_BLOCKING 2
#define ADAPTIVE_STRIPING 3
#define PROCESS_BINDING 4

extern int progress;

#define PARTIAL 0
#define ASYNC 1

extern int sm_scheduling;

/* valid values for s/r scheduling */
#define USE_FIRST 0
#define ROUND_ROBIN 1
#define ALT_ROUND_ROBIN 8
#define SR_BINDING_BLOCK 3
#define SR_BINDING_CYCLIC 5


#endif
/* valid setttings: 4X, 1X, 4X_1X, 1X_4X */
extern int link_speed[MAX_SUBCHANNELS];

extern int striping_threshold;

/* scheduling polices */

/* valid values for RDMA scheduling */
#define RDMA_USE_FIRST 0
#define RDMA_EVEN_STRIPING 1
#define RDMA_WEIGHTED_STRIPING 2
#define RDMA_BINDING_BLOCK 3
#define RDMA_BINDING_CYCLIC 4
#define RDMA_SENDER_DYNAMIC 5

#define DYNAMIC_TOTAL_WEIGHT 1024
extern int rdma_scheduling;

/* valid values for s/r scheduling */
#define SR_USE_FIRST 0
#define SR_ROUND_ROBIN 1
#define SR_BINDING_BLOCK 3
#define SR_BINDING_CYCLIC 5
extern int sr_scheduling;

/* used for policy SR_ROUND_ROBIN */
extern int sr_rr_window;
#define DEFAULT_RR_WINDOW 20

#if 0 
#define D_PRINT(fmt, args...)	{fprintf(stderr, "[%d][%s:%d]", ibv_dev.me, __FILE__, __LINE__);\
				 fprintf(stderr, fmt, ## args); fflush(stderr);}
#else
#define D_PRINT(fmt, args...)
#endif

#if 0
#define V_PRINT(fmt, args...)   {fprintf(stderr, "[%d][%s:%d]", ibv_dev.me, __FILE__, __LINE__);\
                 fprintf(stderr, fmt, ## args); fflush(stderr);}
#else
#define V_PRINT(fmt, args...)
#endif


#if 0 
#define M_PRINT(fmt, args...)   {fprintf(stderr, "[%d][%s:%d]", ibv_dev.me, __FILE__, __LINE__);\
                 fprintf(stderr, fmt, ## args); fflush(stderr);}
#else
#define M_PRINT(fmt, args...)
#endif

#endif   /* _VAPI_HEADER_H */
