/* Copyright (c) 2002-2006, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#include "rdma_fast_path.h"

#define DEFAULT_SUBCHANNEL 0
#ifdef RDMA_FAST_PATH

void init_rdma_fast_path(ibv_connection_t *c)
{
    int j;
    for(j = 0; j < num_subchannels; j++) {
        c->rdma_credit[j] = 0;

#define RDMA_ALIGNMENT 64
#define RDMA_ALIGNMENT_OFFSET (64)
		
		c->RDMA_send_buf_orig[j] = NULL;
		c->RDMA_recv_buf_orig[j] = NULL;
        /* allocate RDMA buffers */
        c->RDMA_send_buf_orig[j] =
            malloc(sizeof(struct vbuf) * viadev_num_rdma_buffer +
                    2 * RDMA_ALIGNMENT);
        c->RDMA_recv_buf_orig[j] =
            malloc(sizeof(struct vbuf) * viadev_num_rdma_buffer +
                    2 * RDMA_ALIGNMENT);
        /* align vbuf->buffer to 64 byte boundary */
			
		D_PRINT("allocating RDMA_send_buf\n");
        c->RDMA_send_buf[j] =
            (struct vbuf*) ((unsigned long) (c->RDMA_send_buf_orig[j])
                           / RDMA_ALIGNMENT * RDMA_ALIGNMENT +
                             RDMA_ALIGNMENT + RDMA_ALIGNMENT_OFFSET);
        c->RDMA_recv_buf[j] =
            (struct vbuf*) ((unsigned long) (c->RDMA_recv_buf_orig[j])
                            / RDMA_ALIGNMENT * RDMA_ALIGNMENT +
                            RDMA_ALIGNMENT + RDMA_ALIGNMENT_OFFSET);

		
		assert(c->RDMA_send_buf[j] != NULL);
		assert(c->RDMA_recv_buf[j] != NULL);
        /* zero buffers */
		memset(c->RDMA_send_buf[j], 0,
                sizeof(struct vbuf) * viadev_num_rdma_buffer);
        memset(c->RDMA_recv_buf[j], 0,
                sizeof(struct vbuf) * viadev_num_rdma_buffer);
        /* set pointers */
        c->phead_RDMA_send[j] = 0;
        c->ptail_RDMA_send[j] = viadev_num_rdma_buffer - 1;
        c->p_RDMA_recv[j] = 0;
        c->p_RDMA_recv_tail[j] = viadev_num_rdma_buffer - 1;


        /* Register the buffers */
        c->RDMA_send_buf_hndl[j] = register_memory(c->RDMA_send_buf[j], 
                sizeof(struct vbuf) * viadev_num_rdma_buffer, j/num_ports);
        c->RDMA_recv_buf_hndl[j] = register_memory(c->RDMA_recv_buf[j],
                sizeof(struct vbuf) * viadev_num_rdma_buffer, j/num_ports);

        if((NULL == c->RDMA_recv_buf_hndl[j]) || 
                (NULL == c->RDMA_send_buf_hndl[j])) {
            error_abort_all(IBV_RETURN_ERR, 
                    "Can't register RDMA buffers\n");
        }
    }
    /* remote address and handle not available yet */
    c->remote_address_received = 0;
}

void finalize_rdma_fast_path(void)
{
    int i, j;
    ibv_connection_t *c;

    for(i = 0; i < ibv_dev.np; i++) {

        c = &ibv_dev.connections[i];

        if(i == ibv_dev.me) {
            continue;
        }
        for(j = 0; j < num_subchannels; j++){
            if(deregister_memory(c->RDMA_send_buf_hndl[j])) {
                error_abort_all(IBV_RETURN_ERR,
                        "could not deregister RDMA buffer\n");
            }
        
            if(deregister_memory(c->RDMA_recv_buf_hndl[j])) {
                error_abort_all(IBV_RETURN_ERR,
                        "could not deregister RDMA buffer\n");
            }
            free(c->RDMA_send_buf_orig[j]);
            free(c->RDMA_recv_buf_orig[j]);
        }
    }
}

void rdma_buffer_address_exchange(void)
{
    int i, j;
    ibv_connection_t *c;
    vbuf *v;
    viadev_packet_rdma_address *h;

    for(i = 0; i < ibv_dev.np; i++) {
        c = &ibv_dev.connections[i];

        if(i == ibv_dev.me) {
            continue;
        }

        for(j = 0; j < num_subchannels; j++) {
            v = get_vbuf();
            h = (viadev_packet_rdma_address *) VBUF_BUFFER_START(v);

            /* set up the packet */
            PACKET_SET_HEADER(h, c, VIADEV_RDMA_ADDRESS);
            h->envelope.src_lrank = i;
            /* set address and handle */
            h->RDMA_address = c->RDMA_recv_buf[j];
            h->RDMA_hndl = c->RDMA_recv_buf_hndl[j]->rkey;

            /* finally send it off */
			h->subchannel = j;

            /* Use the default subchannel to send all the RDMA Addresses */
            vbuf_init_send(v, sizeof(viadev_packet_rdma_address), DEFAULT_SUBCHANNEL);
            viadev_post_send(c, v, DEFAULT_SUBCHANNEL);
        }
    }
    /* get all the address */
    for(i = 0; i < ibv_dev.np; i++) {

        c = &ibv_dev.connections[i];

        if(ibv_dev.me == i) {
            continue;
        }

        while (c->remote_address_received < num_subchannels) {
            MPID_DeviceCheck(1);
        }
    }
}

#endif /* RDMA_FAST_PATH */
