/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */

/* Copyright (c) 2002-2006, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */


#ifndef _VBUF_H
#define _VBUF_H

#include "ibverbs_header.h"

#ifdef _IA32_
    #ifdef _SMALL_CLUSTER
    #define VBUF_TOTAL_SIZE            (1024*12)
    #elif defined (_MEDIUM_CLUSTER)
    #define VBUF_TOTAL_SIZE            (1024*4)
    #elif defined (_LARGE_CLUSTER)
    #define VBUF_TOTAL_SIZE            (1024*2)
    #else 
    #define VBUF_TOTAL_SIZE            (1024*12)
    #endif
#elif defined (_EM64T_)
    #ifdef _SMALL_CLUSTER
    #define VBUF_TOTAL_SIZE            (1024*12)
    #elif defined (_MEDIUM_CLUSTER)
    #define VBUF_TOTAL_SIZE            (1024*4)
    #elif defined (_LARGE_CLUSTER)
    #define VBUF_TOTAL_SIZE            (1024*2)
    #else
    #define VBUF_TOTAL_SIZE            (1024*16)
    #endif
#elif defined (_IA64_)
    #ifdef _SMALL_CLUSTER
    #define VBUF_TOTAL_SIZE            (1024*8)
    #elif defined (_MEDIUM_CLUSTER)
    #define VBUF_TOTAL_SIZE            (1024*4)
    #elif defined (_LARGE_CLUSTER)
    #define VBUF_TOTAL_SIZE            (1024*2)
    #else
    #define VBUF_TOTAL_SIZE            (1024*8)
    #endif
#elif defined (_X86_64_)
    #ifdef _SMALL_CLUSTER
    #define VBUF_TOTAL_SIZE            (1024*12)
    #elif defined (_MEDIUM_CLUSTER)
    #define VBUF_TOTAL_SIZE            (1024*4)
    #elif defined (_LARGE_CLUSTER)
    #define VBUF_TOTAL_SIZE            (1024*2)
    #else
    #define VBUF_TOTAL_SIZE            (1024*12)
    #endif
#elif defined (MAC_OSX)
    #ifdef _SMALL_CLUSTER
    #define VBUF_TOTAL_SIZE            (1024*16)
    #elif defined (_MEDIUM_CLUSTER)
    #define VBUF_TOTAL_SIZE            (1024*8)
    #elif defined (_LARGE_CLUSTER)
    #define VBUF_TOTAL_SIZE            (1024*2)
    #else
    #define VBUF_TOTAL_SIZE            (1024*16)
    #endif
#elif defined (_PPC64_)
    #ifdef _SMALL_CLUSTER
    #define VBUF_TOTAL_SIZE            (1024*16)
    #elif defined (_MEDIUM_CLUSTER)
    #define VBUF_TOTAL_SIZE            (1024*8)
    #elif defined (_LARGE_CLUSTER)
    #define VBUF_TOTAL_SIZE            (1024*2)
    #else
    #define VBUF_TOTAL_SIZE            (1024 * 16)
    #endif
#endif


#define VBUF_ALIGNMENT VBUF_TOTAL_SIZE
#define VBUF_FLAG_TYPE u_int32_t

#if defined(RDMA_FAST_PATH)

#define VBUF_BUFFER_SIZE                                            \
(VBUF_TOTAL_SIZE - sizeof(IBV_DESCRIPTOR) - sizeof(void *)         \
 - sizeof(struct vbuf_region*) -sizeof(int) - sizeof(int) - 	    \
sizeof(VBUF_FLAG_TYPE) - sizeof(VBUF_FLAG_TYPE))

#else
#define VBUF_BUFFER_SIZE                                            \
(VBUF_TOTAL_SIZE - sizeof(IBV_DESCRIPTOR) - sizeof(void *)         \
 - sizeof(struct vbuf_region*) - sizeof(int) - sizeof(int))
#endif

/*
 * brief justification for vbuf format:
 * descriptor must be aligned (64 bytes).
 * vbuf size must be multiple of this alignment to allow contiguous allocation
 * descriptor and buffer should be contiguous to allow via implementations that
 * optimize contiguous descriptor/data (? how likely ?)
 * need to be able to store send handle in vbuf so that we can mark sends
 * complete when communication completes. don't want to store
 * it in packet header because we don't always need to send over the network.
 * don't want to store at beginning or between desc and buffer (see above) so
 * store at end. 
 */

struct ibv_descriptor {
    union {
        struct ibv_recv_wr  rr;
        struct ibv_send_wr  sr;

    } u;
    struct ibv_sge sg_entry;
	int subchannel;
    void *next;
};

typedef struct ibv_descriptor IBV_DESCRIPTOR;

typedef struct vbuf {


#if (defined(RDMA_FAST_PATH))
    VBUF_FLAG_TYPE head_flag;
#endif

    unsigned char buffer[VBUF_BUFFER_SIZE];

#if (defined(RDMA_FAST_PATH))
    uint32_t padding;
#endif

    IBV_DESCRIPTOR desc;

    /* NULL shandle means not send or not complete. Non-null
     * means pointer to send handle that is now complete. Used
     * by viadev_process_send
     */
    void *shandle;
    struct vbuf_region *region;
    uint32_t grank;
	int subchannel;
} vbuf;
/* one for head and one for tail */

#define VBUF_FAST_RDMA_EXTRA_BYTES (2*sizeof(VBUF_FLAG_TYPE))

#define FAST_RDMA_ALT_TAG 0x8000
#define FAST_RDMA_SIZE_MASK 0x7fff

/*
 * Vbufs are allocated in blocks and threaded on a single free list.
 *
 * These data structures record information on all the vbuf
 * regions that have been allocated.  They can be used for
 * error checking and to un-register and deallocate the regions
 * at program termination.
 *
 */
typedef struct vbuf_region {
    struct ibv_mr *mem_handle[MAX_NUM_HCAS];  
    /* memory handle for entire region */

    void *malloc_start;         /* used to free region later */
    void *malloc_end;           /* to bracket mem region */
    int count;                  /* number of vbufs in region */
    struct vbuf *vbuf_head;     /* first vbuf in region */
    struct vbuf_region *next;   /* thread vbuf regions */
} vbuf_region;

/* ack. needed here after vbuf is defined. need to clean up header files
 * and dependencies.
 */

#include "viutil.h"


void dump_vbuf_regions(void);
void dump_vbuf_region(vbuf_region * r);


void allocate_vbufs(void);

void deallocate_vbufs(void);

extern vbuf *free_vbuf_head;

vbuf *get_vbuf(void);
void release_vbuf(vbuf * v);
void vbuf_init_send(vbuf * v, unsigned long len, int subchannel);
void vbuf_init_recv(vbuf * v, unsigned long len, int subchannel);
void vbuf_init_sendrecv(vbuf * v, unsigned long len);

void vbuf_init_sendrecv_immediate(vbuf * v, unsigned long len,
                                  int immediate);

void vbuf_init_rput(vbuf * v, void *local_address,
                    uint32_t lkey, void *remote_address,
                    uint32_t rkey, int len);
#ifdef VIADEV_RGET_SUPPORT
void vbuf_init_rget(vbuf * v,
                    void *local_address,
                    struct ibv_mr* local_memhandle,
                    void *remote_address,
                    struct ibv_mr* remote_memhandle, int nbytes);
#endif

void dump_vbuf(char *msg, vbuf * v);

/* 
 * Macros for working with vbufs 
 */

#define VBUF_BUFFER_START(v) (&v->buffer[0])

#define VBUF_DATA_SIZE(type) (VBUF_BUFFER_SIZE - sizeof(type))


#endif                          /* _VBUF_H */
