/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */

/* Copyright (c) 2002-2006, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "viaparam.h"
#include "vbuf.h"
#include "viapacket.h"
#include "ibverbs_const.h"
#include "ibverbs_header.h"

#ifdef VIADEV_HAVE_RDMA_LIMIT
int viadev_rdma_limit = VIADEV_RDMA_LIMIT;
#endif

uint32_t inline_size;

unsigned long viadev_sq_size = DEFAULT_MAX_SEND_WQE;
/* Max number of entries on the Send Q of QPs per connection.
 *  * Should be about (prepost_depth + extra).
 *   * Must be within NIC MaxQpEntries limit.
 *    * Size will be adjusted below.
 *     */

unsigned long viadev_rq_size;

unsigned long viadev_cq_size = DEFAULT_MAX_CQ_SIZE;

int viadev_default_port = DEFAULT_HCA_PORT;

int viadev_num_rdma_buffer = NUM_RDMA_BUFFER;

int viadev_prepost_depth = PREPOST_DEPTH;
int viadev_prepost_threshold = PREPOST_THRESHOLD;
int viadev_initial_prepost_depth = INITIAL_PREPOST_DEPTH;
int viadev_initial_credits;

static int viadev_vbuf_extra = 10;

int viadev_vbuf_pool_size = VBUF_POOL_SIZE;

int viadev_vbuf_secondary_pool_size = VBUF_SECONDARY_POOL_SIZE;

int viadev_vbuf_max = -1;

int viadev_prepost_rendezvous_extra = PREPOST_RENDEZVOUS_EXTRA;

int viadev_rendezvous_threshold = RENDEZVOUS_THRESHOLD;

int viadev_spin_count = 100;

int viadev_prepost_noop_extra = PREPOST_NOOP_EXTRA;

int viadev_credit_preserve = CREDIT_PRESERVE;

int viadev_credit_notify_threshold = CREDIT_NOTIFY_THRESHOLD;
int viadev_dynamic_credit_threshold = DYNAMIC_CREDIT_THRESHOLD;
int viadev_verbose = 0;

unsigned int        viadev_ndreg_entries = DEFAULT_NDREG_ENTERIES;

unsigned long       viadev_max_rdma_size = DEFAULT_MAX_RDMA_SIZE;

int     viadev_default_mtu = DEFAULT_MTU;
int     viadev_default_qp_ous_rd_atom ;
int     viadev_default_psn = DEFAULT_PSN;
int     viadev_default_pkey_ix;
u_int32_t       viadev_default_p_key ;
int     viadev_default_min_rnr_timer = DEFAULT_MIN_RNR_TIMER;
int     viadev_default_service_level = DEFAULT_SERVICE_LEVEL;
int     viadev_default_time_out ;
int     viadev_default_static_rate = DEFAULT_STATIC_RATE;
u_int8_t        viadev_default_src_path_bits = DEFAULT_SRC_PATH_BITS;
int     viadev_max_fast_eager_size ;
int     viadev_default_retry_count = DEFAULT_RETRY_COUNT;
int     viadev_default_rnr_retry = DEFAULT_RNR_RETRY;
u_int32_t       viadev_default_max_sg_list ;
u_int8_t        viadev_default_max_rdma_dst_ops ;

#ifdef _SMP_
#ifdef _AFFINITY_
/* Affinity is enabled by default */
unsigned int viadev_enable_affinity=1;
#endif
#endif

void viadev_init_parameters(int num_proc, int me)
{

    char *value;
    /* ===============================================================
 	*      * Examine the environment variables to see in any of these
 	*      * were re-defined on the command line.
 	*      * ===============================================================
 	**/
    if ((value = getenv("VIADEV_VERBOSE")) != NULL) {
        viadev_verbose = atoi(value);
    }
    memset(ibv_dev.device_name, 0, 32);

    if ((value = getenv("VIADEV_DEVICE")) != NULL) {
        strncpy(ibv_dev.device_name, value, 32);
    } else {
        strncpy(ibv_dev.device_name, VIADEV_DEFAULT_DEVICE, 32);
    }

    if ((value = getenv("VIADEV_DEFAULT_PORT")) != NULL) {
        viadev_default_port = (int)atoi(value);
    }

    if ((value = getenv("NUM_PORTS")) != NULL) {
        num_phy_ports = atoi(value);
    }
    if ((value = getenv("NUM_QP_PER_PORT")) != NULL) {
        num_qp_per_port = atoi(value);
    }
    if ((value = getenv("NUM_HCAS")) != NULL) {
        num_hcas = atoi(value);
    }
	num_ports = num_phy_ports * num_qp_per_port;
	num_subchannels = num_ports * num_hcas;
    

    if ( (value = getenv("SM_SCHEDULING")) != NULL) {
        if (!strcmp(value, "USE_FIRST")) {
            sm_scheduling = USE_FIRST;
        } else if (!strcmp(value, "ROUND_ROBIN")) {
            sm_scheduling = ROUND_ROBIN;
        } else if (!strcmp(value, "PROCESS_BINDING")) {
            sm_scheduling = PROCESS_BINDING;
        } else {
            fprintf(stderr, "invalid small message scheduling\n");
        }
    } else {
        sm_scheduling = ROUND_ROBIN;
    }

    if ( (value = getenv("LM_SCHEDULING")) != NULL) {
        if (!strcmp(value, "ROUND_ROBIN")) {
            lm_scheduling = ROUND_ROBIN;
        } else if (!strcmp(value, "WEIGHTED_STRIPING")) {
            lm_scheduling = WEIGHTED_STRIPING;
        } else if (!strcmp(value, "EVEN_STRIPING")) {
            lm_scheduling = EVEN_STRIPING;
        } else if (!strcmp(value, "STRIPE_BLOCKING")) {
            lm_scheduling = STRIPE_BLOCKING;
        } else if (!strcmp(value, "ADAPTIVE_STRIPING")) {
            lm_scheduling = ADAPTIVE_STRIPING;
        } else if (!strcmp(value, "PROCESS_BINDING")) {
            lm_scheduling = PROCESS_BINDING;
        } else {
            fprintf(stderr, "invalid large message scheduling\n");
        }
    } else {
        lm_scheduling = STRIPE_BLOCKING;
    }


    /* Since using multiple rails may increase resource usage
     * RDMA credits can be shared amongst the connections, to
     * have the same resource utilization as single-rail implemenation */
    if ((getenv("RESOURCE_SHARING")) != NULL){
            viadev_num_rdma_buffer = viadev_num_rdma_buffer 
											/ num_subchannels;
    }


#ifdef VIADEV_HAVE_RDMA_LIMIT
    if ((value = getenv("VIADEV_RDMA_LIMIT")) != NULL) {
        viadev_rdma_limit = atoi(value);
    }
#endif
    if ((value = getenv("VIADEV_MAX_RDMA_SIZE")) != NULL) {
        viadev_max_rdma_size = (unsigned long)atol(value);
    }

    if ((value = getenv("VIADEV_DEFAULT_QP_OUS_RD_ATOM")) != NULL) {
        viadev_default_qp_ous_rd_atom = (u_int8_t)atoi(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_PSN")) != NULL) {
        viadev_default_psn = (int)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_PKEY_IX")) != NULL) {
        viadev_default_pkey_ix = (int)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_P_KEY")) != NULL) {
        viadev_default_p_key = (u_int32_t)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_MIN_RNR_TIMER")) != NULL) {
        viadev_default_min_rnr_timer = (int)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_SERVICE_LEVEL")) != NULL) {
        viadev_default_service_level = (int)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_TIME_OUT")) != NULL) {
        viadev_default_time_out = (int)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_STATIC_RATE")) != NULL) {
        viadev_default_static_rate = (int)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_SRC_PATH_BITS")) != NULL) {
        viadev_default_src_path_bits = (u_int8_t)atoi(value);
    }
    if ((value = getenv("VIADEV_MAX_FAST_EAGER_SIZE")) != NULL) {
        viadev_max_fast_eager_size = (int)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_RETRY_COUNT")) != NULL) {
        viadev_default_retry_count = (int)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_RNR_RETRY")) != NULL) {
        viadev_default_rnr_retry = (int)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_MAX_SG_LIST")) != NULL) {
        viadev_default_max_sg_list = (u_int32_t)atol(value);
    }
    if ((value = getenv("VIADEV_DEFAULT_MAX_RDMA_DST_OPS")) != NULL) {
        viadev_default_max_rdma_dst_ops = (u_int8_t)atol(value);
    }

    if ((value = getenv("VIADEV_NDREG_ENTRIES")) != NULL) {
        viadev_ndreg_entries = (unsigned int)atoi(value);
    }
    if ((value = getenv("VIADEV_SQ_SIZE")) != NULL) {
        viadev_sq_size = atol(value);
    }
    if ((value = getenv("VIADEV_CQ_SIZE")) != NULL) {
        viadev_cq_size = atol(value);
    }
    if ((value = getenv("VIADEV_NUM_RDMA_BUFFER")) != NULL) {
        viadev_num_rdma_buffer = atoi(value);
        if (viadev_num_rdma_buffer < 0)
            viadev_num_rdma_buffer = 0;
        if (viadev_num_rdma_buffer == 1)
            viadev_num_rdma_buffer = 2; /* need 1 extra */
    }
    if ((value = getenv("VIADEV_PREPOST_DEPTH")) != NULL) {
        viadev_prepost_depth = atoi(value);
    }
    if ((value = getenv("VIADEV_PREPOST_THRESHOLD")) != NULL) {
        viadev_prepost_threshold = atoi(value);
    }
    if ((value = getenv("VIADEV_VBUF_POOL_SIZE")) != NULL) {
    viadev_vbuf_pool_size = atoi(value);
    }
    if ((value = getenv("VIADEV_VBUF_SECONDARY_POOL_SIZE")) != NULL) {
    viadev_vbuf_secondary_pool_size = atoi(value);
    }
    if ((value = getenv("VIADEV_VBUF_EXTRA")) != NULL) {
        viadev_vbuf_extra = atoi(value);
    }
    if ((value = getenv("VIADEV_VBUF_MAX")) != NULL) {
        viadev_vbuf_max = atoi(value);
    }
    if ((value = getenv("VIADEV_PREPOST_RENDEZVOUS_EXTRA")) != NULL) {
        viadev_prepost_rendezvous_extra = atoi(value);
    }
    if ((value = getenv("VIADEV_RENDEZVOUS_THRESHOLD")) != NULL) {
        viadev_rendezvous_threshold = atoi(value);
    }
    if ((value = getenv("VIADEV_SPIN_COUNT")) != NULL) {
        viadev_spin_count = atoi(value);
    }
    if ((value = getenv("VIADEV_PREPOST_NOOP_EXTRA")) != NULL) {
        viadev_prepost_noop_extra = atoi(value);
    }
    if ((value = getenv("VIADEV_INITIAL_PREPOST_DEPTH")) != NULL) {
        viadev_initial_prepost_depth = atoi(value);
    }
    if (viadev_initial_prepost_depth >
        viadev_prepost_depth + viadev_prepost_noop_extra)
        viadev_initial_prepost_depth =
            viadev_prepost_depth + viadev_prepost_noop_extra;
    if (viadev_initial_prepost_depth <= 0)
        viadev_initial_prepost_depth = 1;
    if (viadev_initial_prepost_depth <= viadev_prepost_noop_extra) {
        viadev_initial_credits = viadev_initial_prepost_depth;
    } else {
        viadev_initial_credits =
            viadev_initial_prepost_depth - viadev_prepost_noop_extra;
    }
    if ((value = getenv("VIADEV_CREDIT_PRESERVE")) != NULL) {
        viadev_credit_preserve = atoi(value);
    }
    if ((value = getenv("VIADEV_CREDIT_NOTIFY_THRESHOLD")) != NULL) {
        viadev_credit_notify_threshold = atoi(value);
    }
    if ((viadev_prepost_depth - viadev_prepost_threshold) <=
        MAX(viadev_credit_notify_threshold,
            viadev_dynamic_credit_threshold)) {
        viadev_prepost_threshold =
            viadev_prepost_depth - 1 - MAX(viadev_credit_notify_threshold,
                                           viadev_dynamic_credit_threshold);
    }
    if (viadev_prepost_threshold < 0)
        viadev_prepost_threshold = 0;

#ifdef _SMP_
    /* Should set in KBytes */
    if ( (value = getenv("SMP_EAGERSIZE")) != NULL ) {
        smp_eagersize = atoi(value);
    }
    /* Should set in MBytes */
    if ( (value = getenv("SMPI_LENGTH_QUEUE")) != NULL ) {
        smpi_length_queue = atoi(value);
    }
#ifdef _AFFINITY_
    if ((value = getenv("VIADEV_ENABLE_AFFINITY")) != NULL) {
            viadev_enable_affinity = atoi(value);
    }
#endif

#endif

    /* compute Receive Q size */
    viadev_rq_size =
        viadev_prepost_depth + viadev_prepost_rendezvous_extra +
        viadev_prepost_noop_extra;

}

void dump_param_values(FILE * fd, int me)
{
    char *int_fmt = "%3d: %32s = %d\n";

    fprintf(fd, int_fmt, me, "VIADEV_VERBOSE", viadev_verbose);

#ifdef VIADEV_HAVE_RDMA_LIMIT
    fprintf(fd, int_fmt, me, "VIADEV_RDMA_LIMIT", viadev_rdma_limit);
#endif

    fprintf(fd, int_fmt, me, "VIADEV_MAX_RDMA_SIZE", (int)viadev_max_rdma_size);

    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_MTU", (int)viadev_default_mtu);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_QP_OUS_RD_ATOM", (int)viadev_default_qp_ous_rd_atom);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_PSN", (int)viadev_default_psn);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_PKEY_IX", (int)viadev_default_pkey_ix);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_P_KEY", (int)viadev_default_p_key);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_MIN_RNR_TIMER", (int)viadev_default_min_rnr_timer);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_SERVICE_LEVEL", (int)viadev_default_service_level);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_TIME_OUT", (int)viadev_default_time_out);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_STATIC_RATE", (int)viadev_default_static_rate);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_SRC_PATH_BITS", (int)viadev_default_src_path_bits);
    fprintf(fd, int_fmt, me, "VIADEV_MAX_FAST_EAGER_SIZE", (int)viadev_max_fast_eager_size);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_RETRY_COUNT", (int)viadev_default_retry_count);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_RNR_RETRY", (int)viadev_default_rnr_retry);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_MAX_SG_LIST", (int)viadev_default_max_sg_list);
    fprintf(fd, int_fmt, me, "VIADEV_DEFAULT_MAX_RDMA_DST_OPS", (int)viadev_default_max_rdma_dst_ops);


    fprintf(fd, int_fmt, me, "VIADEV_NDREG_ENTRIES", viadev_ndreg_entries);
    fprintf(fd, int_fmt, me, "VIADEV_SQ_SIZE", (int) viadev_sq_size);
    fprintf(fd, int_fmt, me, "VIADEV_CQ_SIZE", (int) viadev_cq_size);
    fprintf(fd, int_fmt, me, "VIADEV_NUM_RDMA_BUFFER",
            viadev_num_rdma_buffer);
    fprintf(fd, int_fmt, me, "VIADEV_PREPOST_DEPTH", viadev_prepost_depth);
    fprintf(fd, int_fmt, me, "VIADEV_INITIAL_PREPOST_DEPTH",
            viadev_initial_prepost_depth);
    fprintf(fd, int_fmt, me, "VIADEV_INITIAL_PREPOST_DEPTH",
            viadev_initial_prepost_depth);
    fprintf(fd, int_fmt, me,
            "VIADEV_PREPOST_THRESHOLD", viadev_prepost_threshold);
    fprintf(fd, int_fmt, me,
            "VIADEV_VBUF_POOL_SIZE", viadev_vbuf_pool_size);
    fprintf(fd, int_fmt, me,
            "VIADEV_VBUF_SECONDARY_POOL_SIZE",
            viadev_vbuf_secondary_pool_size);
    fprintf(fd, int_fmt, me, "VIADEV_VBUF_MAX", viadev_vbuf_max);
    fprintf(fd, int_fmt, me,
            "VIADEV_PREPOST_RENDEZVOUS_EXTRA",
            viadev_prepost_rendezvous_extra);
    fprintf(fd, int_fmt, me,
            "VIADEV_PREPOST_NOOP_EXTRA", viadev_prepost_noop_extra);
    fprintf(fd, int_fmt, me, "VIADEV_CREDIT_PRESERVE",
            viadev_credit_preserve);
    fprintf(fd, int_fmt, me, "VIADEV_CREDIT_NOTIFY_THRESHOLD",
            viadev_credit_notify_threshold);
    fprintf(fd, int_fmt, me, "VIADEV_DYNAMIC_CREDIT_THRESHOLD",
            viadev_dynamic_credit_threshold);
    fprintf(fd, int_fmt, me, "VIADEV_RENDEZVOUS_THRESHOLD",
            viadev_rendezvous_threshold);
    fprintf(fd, int_fmt, me, "VIADEV_SPIN_COUNT", viadev_spin_count);

}
