/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */

/* Copyright (c) 2002-2006, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */
/*
 * viaparam.h 
 * 
 * This file contains "tunable parameters" for MVICH. A lot of work needs to be
 * done to find out what are the right values, and in fact whether this is 
 * the right set of knobs. Many are settable from the command line and
 * passed to the individual processes as environment variables.
 * Default value are set in the viaparam.c
 */

#ifndef _VIAPARAM_H
#define _VIAPARAM_H

#include "ibverbs_header.h"

#include <stdio.h>

#define MIN(a,b) ((a)<(b)?(a):(b))
#define MAX(a,b) ((a)>(b)?(a):(b))

extern uint32_t             inline_size;

#ifdef _SMP_
extern int  smp_eagersize; /* EAGER threshold in SMP path */
extern int  smpi_length_queue; /* Shared memory buffer pool size */

#ifdef _AFFINITY_
extern unsigned int         viadev_enable_affinity;
#endif

#endif

extern u_int32_t        viadev_default_max_sg_list;
extern u_int8_t         viadev_default_max_rdma_dst_ops;
extern unsigned int         viadev_ndreg_entries;
extern unsigned long        viadev_sq_size;
extern unsigned long        viadev_rq_size;

extern unsigned long        viadev_cq_size;
extern int          viadev_default_port;

extern int viadev_num_rdma_buffer;
extern int viadev_prepost_depth;
extern int viadev_prepost_threshold;
extern int viadev_initial_prepost_depth;
extern int viadev_initial_credits;
extern int viadev_vbuf_pool_size;
extern int viadev_vbuf_secondary_pool_size;
extern int viadev_vbuf_max;
extern int viadev_spin_count;
extern int viadev_rendezvous_threshold;
extern int viadev_prepost_rendezvous_extra;
extern int viadev_prepost_noop_extra;
extern int viadev_credit_preserve;
extern int viadev_credit_notify_threshold;
extern int viadev_dynamic_credit_threshold;
extern unsigned long        viadev_max_rdma_size;

extern int viadev_default_static_rate;
void viadev_init_parameters(int num_proc, int me);
void dump_param_values(FILE * fd, int me);

#ifdef VIADEV_HAVE_RDMA_LIMIT
#error Cannot define macro VIADEV_HAVE_RDMA_LIMIT use \
    environment variable VIADEV_SQ_SIZE to control \
    max number of outstanding RDMA
#endif

#endif                          /* _VIAPARAM_H */
