#ifndef _VIAPRIV_H
#define _VIAPRIV_H

#include "ibverbs_header.h"
#include "vbuf.h"
#include "viapacket.h"
#include "req.h"
#include "dreg.h"

#if defined(RDMA_FAST_PATH)

#define MAX_NO_COMPLETION 1

#define PRIMARY_FLAG (7777777)
#define SECONDARY_FLAG (88888888)

#define NORMAL_VBUF_FLAG (222)
#define RPUT_VBUF_FLAG (333)
#define VBUF_FLAG_TYPE u_int32_t

#define FREE_FLAG (0)
#define BUSY_FLAG (1)

#define IS_BUSY_VBUF(v) (v->padding != FREE_FLAG)

#define IS_NORMAL_VBUF(v) (v->padding == NORMAL_VBUF_FLAG)

#define IS_RPUT_VBUF(v) (v->padding == RPUT_VBUF_FLAG)

#endif


#define NIC_0 0
#define SUBCHANNEL_0 0

/*
 * This structure is used to implement a vbuf backlog queue
 * for posting packet sends.  If no remote credits are available
 * on a connection at the time of the packet send, it will be
 * put on the end of the queue.  As credits become available
 * the packets on the queue are sent in the order received.
 */

typedef struct _ibv_backlog_queue_t {
    int                 len;
    /* length of backlog queue */
    vbuf                *vbuf_head;
    /* head of backlog queue */
    vbuf                *vbuf_tail;
    /* tail of backlog queue */

}ibv_backlog_queue_t;

/* Connection data structure. Each one of this
 * per process
 */

typedef struct _ibv_connection_t {
    struct ibv_qp       *qp_hndl[MAX_SUBCHANNELS];
    /* My QP handle for connection to this process */

    uint32_t            global_rank;
    /* Global rank of this process */

    uint32_t            remote_credit[MAX_SUBCHANNELS];
    /* how many vbufs I can consume on remote end. */

    uint32_t            local_credit[MAX_SUBCHANNELS];
    /* accumulate vbuf credit locally here */

    uint32_t            preposts[MAX_SUBCHANNELS];
    /* number of vbufs currently preposted */

    uint32_t            initialized[MAX_SUBCHANNELS];
    /* have all the initial buffers preposted */

    uint32_t            send_wqes_avail[MAX_SUBCHANNELS];
    /* send Q WQES available */

    struct vbuf         *ext_sendq_head[MAX_SUBCHANNELS];
    /* head of queued sends */

    struct vbuf         *ext_sendq_tail[MAX_SUBCHANNELS];
    /* tail queued sends */

    uint16_t   next_packet_expected;
    /* for sequencing (starts at 1) */

    uint16_t   next_packet_to_send;
    /* for sequencing (starts at 1) */

    uint32_t            remote_cc[MAX_SUBCHANNELS];
    /* current guess at remote side's credit count */

    ibv_backlog_queue_t backlog[MAX_SUBCHANNELS];
    /* the backlog queue for this connection */

    uint32_t            rendezvous_packets_expected;
    /* This field is used for managing preposted receives. It is the
     * number of rendezvous packets (r3/rput) expected to arrive for
     * receives we've ACK'd but have not been completed. In general we
     * can't prepost all of these vbufs, but we do prepost extra ones
     * to allow sender to keep the pipe full. As packets come in this
     * field is decremented.  We know when to stop preposting extra
     * buffers when this number goes to zero.
     */

    MPIR_SHANDLE        *shandle_head;
    /* head of queue of send handles to process */

    MPIR_SHANDLE        *shandle_tail;
    /* tail of queue of send handles to process */
    /* these fields are used to remember data transfer operations
     * that are currently in progress on this connection. The
     * send handle list is a queue of send handles representing
     * in-progress rendezvous transfers. It is processed in FIFO
     * order (because of MPI ordering rules) so there is both a head
     * and a tail.
     */

    MPIR_RHANDLE        *rhandle;
    /* current eager receive "in progress" */
    /* The receive handle is a pointer to a single
     * in-progress eager receive. We require that an eager sender
     * send *all* packets associated with an eager receive before
     * sending any others, so when we receive the first packet of
     * an eager series, we remember it by caching the rhandle
     * on the connection.
     */

    struct _ibv_connection_t *nextflow;
    uint32_t            inflow;
    /* these two fields are used *only* by MPID_DeviceCheck to
     * build up a list of connections that have received new
     * flow control credit so that pending operations should be
     * pushed. nextflow is a pointer to the next connection on the
     * list, and inflow is 1 (true) or 0 (false) to indicate whether
     * the connection is currently on the flowlist. This is needed
     * to prevent a circular list.
     */

    uint32_t            barrier_id;
    /* used to distinguish which VIA barrier synchronozations have
     * completed on this connection.  Currently, only used during
     * process teardown.
     */
    uint32_t s_weight[MAX_SUBCHANNELS];
    /* flag indicating we have received remote RDMA memory
     *      * address and handle for this connection */
        /* out-of-order queue for this connection */
    vbuf *ofo_head;
              /* used for binding */
    uint32_t binding_subchannel;
    uint32_t sr_binding_subchannel;

#ifdef RDMA_FAST_PATH
    void                *RDMA_send_buf_orig[MAX_SUBCHANNELS];
    void                *RDMA_recv_buf_orig[MAX_SUBCHANNELS];
    /* Pointers to original RDMA send & recv buffers.
     * These buffers are aligned, so need the original
     * pointers to free the whole buffer */

    struct vbuf         *RDMA_send_buf[MAX_SUBCHANNELS];
    /* RDMA buffers for sending */

    struct vbuf         *RDMA_recv_buf[MAX_SUBCHANNELS];
    /* RDMA buffers for receive */

    struct ibv_mr       *RDMA_send_buf_hndl[MAX_SUBCHANNELS];
    struct ibv_mr       *RDMA_recv_buf_hndl[MAX_SUBCHANNELS];
    /* RDMA buffers needs to be registered */

    uint32_t            remote_RDMA_buf_hndl[MAX_SUBCHANNELS];
    struct vbuf         *remote_RDMA_buf[MAX_SUBCHANNELS];
    /* RDMA buffer on the remote side */

    uint32_t            rdma_credit[MAX_SUBCHANNELS];
    /* current flow control credit accumulated for remote side */

    uint32_t            phead_RDMA_send[MAX_SUBCHANNELS];
    /* Index of RDMA buffer to be used for sending */

    uint32_t            ptail_RDMA_send[MAX_SUBCHANNELS];
    /* Index of RDMA buffer to be used for receiving */
    /* No RDMA available if head == tail */

    uint32_t            p_RDMA_recv[MAX_SUBCHANNELS];
    /* Poll here for incoming message */

    uint32_t            p_RDMA_recv_tail[MAX_SUBCHANNELS];
    /* last free buffer at the receive side */
    uint32_t             num_no_completion[MAX_SUBCHANNELS];
    uint32_t            remote_address_received;

                                   

#if !defined(DISABLE_HEADER_CACHING)
    /* cache part of the packet header to reduce eager message size */
    viadev_packet_eager_start cached_outgoing[MAX_SUBCHANNELS];
    viadev_packet_eager_start cached_incoming[MAX_SUBCHANNELS];
    uint32_t            cached_hit;
    uint32_t            cached_miss;
#endif /* !DISABLE_HEADER_CACHING */
#endif /* RDMA_FAST_PATH */
} ibv_connection_t;

typedef struct {

    struct ibv_device   *nic[MAX_NUM_HCAS];
    /* single NIC */

    struct ibv_context  *context[MAX_NUM_HCAS];
    /* HCA context */

    struct ibv_pd       *ptag[MAX_NUM_HCAS];
    /* single protection tag for all memory registration */

    struct ibv_qp       **qp_hndl;
    /* Array of QP handles for all connections */

    struct ibv_cq       *cq_hndl[MAX_NUM_HCAS];
    /* One CQ for both send & recv */

    uint16_t            my_hca_lid[MAX_SUBCHANNELS];
    /* my HCA LID */

    uint16_t            *lid_table;
    /* Store HCA LID of all processes */

    uint32_t            *qp_table;
    /* Store HCA QP num of all processes */

    unsigned long       maxtransfersize;
    /* max send/rdma size */

    uint32_t            np;
    /* number of processes total */

    uint32_t            me;
    /* my process rank */

    char                *my_name;
    /* A string equivalent of node name */

    uint32_t            global_id;
    /* global id of this parallel app */

    ibv_connection_t    *connections;
    /* array of connections to other processes */

    uint32_t            barrier_id;
    /* Used for barrier operations */

#ifdef RDMA_FAST_PATH
    char                *RDMA_bufs;
    struct ibv_mr       *RDMA_buf_hndl;
#if !defined(DISABLE_HEADER_CACHING)
    viadev_packet_eager_start match_hdr;
#endif  /* !DISABLE_HEADER_CACHING */
#endif  /* RDMA_FAST_PATH */

    uint32_t            *pids;
    char                **processes;
    char                execname[256];
    char        device_name[32]; 
} ibv_info_t;

extern ibv_info_t ibv_dev;

extern char nullsbuffer, nullrbuffer;
int choose_sr_subchannel(ibv_connection_t *c);
int choose_lm_subchannel(ibv_connection_t *c);

/* function prototypes for communication scheduling,
 *    used when multiple channels per connection are used */

/* striping rdma messages across multiple subchannels */
void schedule_rdma(ibv_connection_t *c, char *addr, int size,
        char** stripe_start_addr, int *stripe_size, int do_stripe);
void schedule_rdma_even_striping(ibv_connection_t *c, char *addr, 
		int size, char** stripe_start_addr, int *stripe_size, int do_stripe);
void schedule_rdma_stripe_blocking(ibv_connection_t *c, char *addr, int size,
                                char** stripe_start_addr, int *stripe_size, int do_stripe);
void schedule_rdma_round_robin(ibv_connection_t *c, char *addr, int size,
                    char** stripe_start_addr, int *stripe_size, int do_stripe);
void schedule_rdma_weighted_striping(ibv_connection_t *c, char *addr, int size,
                    char** stripe_start_addr, int *stripe_size, int do_stripe);

void adjust_weights(ibv_connection_t *c, double start_time,
                        double  finish_time[MAX_SUBCHANNELS],
                        double  init_weight[MAX_SUBCHANNELS]);


/* handling out-of-order messages for multiple subchannels */
int check_ofo_queues(vbuf **vbuf_addr);
void enqueue_ofo_queue(vbuf *vbuf_addr, ibv_connection_t *c);
void remove_ofo_queue(vbuf *vbuf_addr, ibv_connection_t *c);
void init_ofo_queues(void);


struct ibv_mr * register_memory(void *buf, int len, int hca);

int deregister_memory(struct ibv_mr* mr);

void viadev_init_backlog_queue(ibv_connection_t * c);

void viadev_post_recv(ibv_connection_t * c, vbuf * v, int subchannel);

void viadev_ext_sendq_queue(ibv_connection_t * c, vbuf * v, int subchannel);
void viadev_post_send(ibv_connection_t * c, vbuf * v, int subchannel);

void viadev_send_noop_ifneeded(ibv_connection_t * c, int subchannel);

void viadev_ext_sendq_send(ibv_connection_t * c, int subchannel);

void viadev_prepost_for_rendezvous(ibv_connection_t * c,
                                   int vbufs_expected,
                                   int subchannel);

void viadev_backlog_send(ibv_connection_t * c, int subchannel);

int viadev_calculate_vbufs_expected(int nbytes, 
        viadev_protocol_t protocol);

void release_recv_rdma(ibv_connection_t * c, vbuf * v);

void viadev_recv_r3(MPIR_RHANDLE * rhandle);

void viadev_recv_rget(MPIR_RHANDLE * rhandle);

void viadev_recv_rput(MPIR_RHANDLE * rhandle);

void viadev_register_recvbuf_if_possible(MPIR_RHANDLE * rhandle);

void viadev_incoming_rendezvous_reply(vbuf * v, 
        ibv_connection_t * c, 
        viadev_packet_rendezvous_reply * header);

void viadev_rendezvous_reply(MPIR_RHANDLE * rhandle);

int fast_rdma_ok(ibv_connection_t * c, int data_size,
        int is_eager_packet, int subchannel);

void post_fast_rdma_with_completion(ibv_connection_t * c, int len,
            int subchannel);

int viadev_eager_ok(int len, ibv_connection_t * c, int subchannel);
void viadev_send_noop(ibv_connection_t * c, int subchannel);

void viadev_rput(ibv_connection_t * c, vbuf * v, void *local_address,
                 uint32_t lkey, void *remote_address,
                 uint32_t rkey, int nbytes);
void viadev_rput_finish(MPIR_SHANDLE * s);

void viadev_post_rdmawrite(ibv_connection_t * c, vbuf * v, 
            int subchannel);

void viadev_incoming_rput_finish(vbuf * v, ibv_connection_t * c,
                                 viadev_packet_rput_finish * h);


#ifdef RDMA_FAST_PATH
#ifndef DISABLE_HEADER_CACHING

int search_header_cache(ibv_connection_t * c,
                        viadev_packet_eager_start * h,
                        int subchannel);
#endif
#endif

/* 
 * on 32 bit machines, we identify (for now) a handle by its address, using the 
 * address for immediate data. 
 * on 64 bit machines, we need to use the pointer index stuff since we don't
 * have 64 bits of immediate data. 
 * We will also need the index stuff when we implement reliability. 
 */

#define REQ_TO_ID(h) ((request_id_t)(h))
#define ID_TO_REQ(id)((union MPIR_HANDLE *)(id))

#define PREPOST_VBUF_RECV(c, subchannel)  {                                     \
    vbuf *v = get_vbuf();                                           \
    vbuf_init_recv(v, VBUF_BUFFER_SIZE, subchannel);                            \
    viadev_post_recv(c, v, subchannel);                                         \
    c->local_credit[subchannel]++;                                              \
    c->preposts[subchannel]++;                                                  \
}

#define SEND_COMPLETE(s) {                                          \
    s->is_complete = 1;                                             \
    if (s->dreg_entry != NULL) {                                    \
        dreg_decr_refcount(s->dreg_entry);                          \
    }                                                               \
    if (s->finish != NULL) {                                        \
        s->finish(s);                                               \
    }                                                               \
    s->dreg_entry = NULL;                                           \
    if (s->ref_count == 0) {                                        \
        switch (s->handle_type) {                                   \
            case MPIR_SEND:                                         \
            {                                                       \
                MPID_SendFree(s);                                   \
                break;                                              \
            }                                                       \
            case MPIR_PERSISTENT_SEND:                              \
            {                                                       \
                MPID_PSendFree(s);                                  \
                break;                                              \
            }                                                       \
            default:                                                \
                error_abort_all(GEN_EXIT_ERR, "SEND_COMPLETE invalid type\n");    \
        }                                                           \
    }                                                               \
}

#define RECV_COMPLETE(r) {                                          \
    r->is_complete = 1;                                             \
    r->s.MPI_ERROR = MPI_SUCCESS;                                   \
    r->s.count = r->len;                                            \
    if (r->dreg_entry != NULL) {                                    \
        dreg_decr_refcount(r->dreg_entry);                          \
    }                                                               \
    r->dreg_entry = NULL;                                           \
    if (r->finish != NULL) {                                        \
        r->finish(r);                                               \
    }                                                               \
    if (r->ref_count == 0) {                                        \
        switch (r->handle_type) {                                   \
            case MPIR_RECV:                                         \
            {                                                       \
                MPID_RecvFree(r);                                   \
                break;                                              \
            }                                                       \
            case MPIR_PERSISTENT_RECV:                              \
            {                                                       \
                MPID_PRecvFree(r);                                  \
                break;                                              \
            }                                                       \
            default:                                                \
                error_abort_all(GEN_EXIT_ERR, "RECV_COMPLETE invalid type\n");    \
        }                                                           \
    }                                                               \
}


#endif /* _VIAPRIV_H */
