/* Copyright (c) 2002-2006, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

/*
 * udapl_const.h 
 * 
 * This file contains "tunable parameters" for VAPI. A lot of work 
 * needs to be done to find out what are the right values, and in 
 * fact whether this is the right set of knobs. Many are settable 
 * from the command line and
 * passed to the individual processes as environment variables.
 */

#ifndef _UDAPL_CONST_H
#define _UDAPL_CONST_H

/*#define CO_PRINT */
#ifdef CO_PRINT
#define co_print(arg...)  \
do {                        \
   printf("[%d]", viadev.me);  \
   printf(arg); \
} while (0)
#else
#define co_print(arg...)
#endif

#ifndef TRUE
#define TRUE 1
#endif

/*
#define NUM_ADDR_BITS           (32)
#define MASK_ADDR_BITS          (0xffffffff)
*/
/* The length of QP representation in characters, could be 4 HEX */
#define QPLEN_XDR               (8)

#define VIADEV_DEFAULT_MTU             (1024)
#define VIADEV_DEFAULT_MTU_SIZE_IBA            (2048)
#define VIADEV_DEFAULT_MTU_SIZE_GM             (4096)
#define VIADEV_DEFAULT_MTU_SIZE_CCIL           (65535)
#define VIADEV_DEFAULT_MTU_SIZE_SOLARIS        (2048)

#define VIADEV_MAX_RDMA_SIZE		(1048576)  

#define VIADEV_DEFAULT_TIME_OUT        (100000)
#define VIADEV_DEFAULT_PROVIDER        "ib0"
#define VIADEV_DEFAULT_ASYNC_EVD_SIZE  (8) 

#define VIADEV_MAX_FAST_EAGER_SIZE 255

#define VIADEV_DEFAULT_MAX_SG_LIST	(20)
#define VIADEV_MAX_WQE                  (300)
#define VIADEV_MAX_WQE_IBA        (300)
#define VIADEV_MAX_WQE_CCIL       (200)
#define VIADEV_MAX_WQE_GM         (8)
#define VIADEV_MAX_WQE_SOLARIS    (300)

#define DAPL_GEN2_MAX_MSG_SIZE               (2147483648)
#define DAPL_DEFAULT_MAX_RDMA_IN             (4)
#define DAPL_DEFAULT_MAX_RDMA_OUT            (4)

#define HOSTNAME_LEN (255)

#ifdef _IA32_
    #ifdef _SMALL_CLUSTER
    #define VIADEV_NUM_RDMA_BUFFER            (32)
    #define DEFAULT_RENDEZVOUS_THRESHOLD      (VBUF_BUFFER_SIZE - \
            sizeof(viadev_packet_eager_start))
    #elif defined (_MEDIUM_CLUSTER)
    #define VIADEV_NUM_RDMA_BUFFER            (16)
    #define DEFAULT_RENDEZVOUS_THRESHOLD      (16384)
    #elif defined (_LARGE_CLUSTER)
    #define VIADEV_NUM_RDMA_BUFFER            (16)
    #define DEFAULT_RENDEZVOUS_THRESHOLD      (16384)
    #else
    #define VIADEV_NUM_RDMA_BUFFER            (32)
    #define DEFAULT_RENDEZVOUS_THRESHOLD      (VBUF_BUFFER_SIZE - \
            sizeof(viadev_packet_eager_start))
    #endif
#elif defined (_EM64T_)
    #ifdef _SMALL_CLUSTER
    #define VIADEV_NUM_RDMA_BUFFER            (32)
    #define DEFAULT_RENDEZVOUS_THRESHOLD      (VBUF_BUFFER_SIZE - \
            sizeof(viadev_packet_eager_start))
    #elif defined (_MEDIUM_CLUSTER)
    #define VIADEV_NUM_RDMA_BUFFER            (16)
    #define DEFAULT_RENDEZVOUS_THRESHOLD      (VBUF_BUFFER_SIZE - \
            sizeof(viadev_packet_eager_start))
    #elif defined (_LARGE_CLUSTER)
    #define VIADEV_NUM_RDMA_BUFFER            (8)
    #define DEFAULT_RENDEZVOUS_THRESHOLD      (8192)
    #else
    #define VIADEV_NUM_RDMA_BUFFER            (32)
    #define DEFAULT_RENDEZVOUS_THRESHOLD      (VBUF_BUFFER_SIZE - \
            sizeof(viadev_packet_eager_start))
    #endif
#elif defined (_IA64_)
    #ifdef _SMALL_CLUSTER
    #define VIADEV_NUM_RDMA_BUFFER            (32)
    #define DEFAULT_RENDEZVOUS_THRESHOLD      (VBUF_BUFFER_SIZE - \
            sizeof(viadev_packet_eager_start))
    #elif defined (_MEDIUM_CLUSTER)
    #define VIADEV_NUM_RDMA_BUFFER            (16)
    #define DEFAULT_RENDEZVOUS_THRESHOLD      (16384)
    #elif defined (_LARGE_CLUSTER)
    #define VIADEV_NUM_RDMA_BUFFER            (16)
    #define DEFAULT_RENDEZVOUS_THRESHOLD      (8192)
    #else
    #define VIADEV_NUM_RDMA_BUFFER            (32)
    #define DEFAULT_RENDEZVOUS_THRESHOLD      (VBUF_BUFFER_SIZE - \
            sizeof(viadev_packet_eager_start))
    #endif
#elif defined (_X86_64_)
    #ifdef _SMALL_CLUSTER
    #define VIADEV_NUM_RDMA_BUFFER            (32)
    #define DEFAULT_RENDEZVOUS_THRESHOLD      (VBUF_BUFFER_SIZE - \
            sizeof(viadev_packet_eager_start))
    #elif defined (_MEDIUM_CLUSTER)
    #define VIADEV_NUM_RDMA_BUFFER            (16)
    #define DEFAULT_RENDEZVOUS_THRESHOLD      (16384)
    #elif defined (_LARGE_CLUSTER)
    #define VIADEV_NUM_RDMA_BUFFER            (16)
    #define DEFAULT_RENDEZVOUS_THRESHOLD      (8192)
    #else
    #define VIADEV_NUM_RDMA_BUFFER            (32)
    #define DEFAULT_RENDEZVOUS_THRESHOLD      (VBUF_BUFFER_SIZE - \
            sizeof(viadev_packet_eager_start))
    #endif
#elif defined (MAC_OSX)
    #ifdef _SMALL_CLUSTER
    #define VIADEV_NUM_RDMA_BUFFER            (128)
    #define DEFAULT_RENDEZVOUS_THRESHOLD      (VBUF_BUFFER_SIZE - \
            sizeof(viadev_packet_eager_start))
    #elif defined (_MEDIUM_CLUSTER)
    #define VIADEV_NUM_RDMA_BUFFER            (64)
    #define DEFAULT_RENDEZVOUS_THRESHOLD      (16384)
    #elif defined (_LARGE_CLUSTER)
    #define VIADEV_NUM_RDMA_BUFFER            (32)
    #define DEFAULT_RENDEZVOUS_THRESHOLD      (8192)
    #else
    #define VIADEV_NUM_RDMA_BUFFER            (128)
    #define DEFAULT_RENDEZVOUS_THRESHOLD      (VBUF_BUFFER_SIZE - \
            sizeof(viadev_packet_eager_start))
    #endif
#elif defined (SOLARIS)
    #ifdef _SMALL_CLUSTER
    #define VIADEV_NUM_RDMA_BUFFER            (32)
    #define DEFAULT_RENDEZVOUS_THRESHOLD      (VBUF_BUFFER_SIZE - \
            sizeof(viadev_packet_eager_start))
    #elif defined (_MEDIUM_CLUSTER)
    #define VIADEV_NUM_RDMA_BUFFER            (16)
    #define DEFAULT_RENDEZVOUS_THRESHOLD      (16384)
    #elif defined (_LARGE_CLUSTER)
    #define VIADEV_NUM_RDMA_BUFFER            (16)
    #define DEFAULT_RENDEZVOUS_THRESHOLD      (16384)
    #else
    #define VIADEV_NUM_RDMA_BUFFER            (32)
    #define DEFAULT_RENDEZVOUS_THRESHOLD      (VBUF_BUFFER_SIZE - \
            sizeof(viadev_packet_eager_start))
    #endif
#endif

#define VIADEV_PREPOST_DEPTH 64
#define VIADEV_VBUF_MIN_POOL_SIZE 100
#define VIADEV_PREPOST_RENDEZVOUS_EXTRA VIADEV_PREPOST_DEPTH
#define DEFAULT_MAX_CQ_SIZE (6000)


#endif                          /* _UDAPL_CONST_H */
